from ..nn.core import Variable


class Regularizer:
    def __call__(self, x):
        return 0.


class L1L2(Regularizer):
    def __init__(self, l1: float = 0., l2: float = 0.):
        self.l1 = Variable(l1, requires_grad=False)
        self.l2 = Variable(l2, requires_grad=False)

    def __call__(self, x: Variable):
        regularization = 0.
        if self.l1:
            regularization += self.l1 * x.abs().sum()
        if self.l2:
            regularization += self.l2 * (x ** 2).sum()
        return regularization


def l1(l=0.01):
    return L1L2(l1=l)


def l2(l=0.01):
    return L1L2(l2=l)


def l1_l2(l1=0.01, l2=0.01):
    return L1L2(l1=l1, l2=l2)



from ..nn.core import Variable
from ..nn.grad_fn import MeanSquaredbackward, MeanAbsolutebackward, BinaryCrossEntropyBackward, SparseCrossEntropyBackward, CrossEntropyBackward
import copy
from ..nn.functional import softmax
from functools import reduce
from ..nn import global_graph as GlobalGraph


class Objective:
    def __init__(self):
        self.in_bounds = []
        self.grad_fn = None

    def __call__(self, y_pred: Variable, y_true: Variable):
        if GlobalGraph.outputs is None:
            GlobalGraph.outputs = y_pred

        return self.forward(y_pred, y_true)

    def forward(self, y_pred: Variable, y_true: Variable) -> Variable:
        raise NotImplemented

    def acc(self, y_pred: Variable, y_true: Variable):
        raise NotImplemented

    def metric(self, y_pred: Variable, y_true: Variable):
        acc = self.acc(y_pred, y_true)
        loss = self.forward(y_pred, y_true)
        return acc, loss.data.tolist()


class MeanSquaredError(Objective):
    def __init__(self):
        super().__init__()
        self.grad_fn = MeanSquaredbackward

    def acc(self, y_pred: Variable, y_true: Variable):
        return 0.

    def forward(self, y_pred: Variable, y_true: Variable):
        loss_val = 0.5 * GlobalGraph.np.sum(GlobalGraph.np.power(y_pred.data - y_true.data, 2)) / y_pred.shape[0]
        loss = Variable(data=loss_val, in_bounds=[y_pred, y_true])
        loss.grad_fn = self.grad_fn
        return loss


class MeanAbsoluteError(Objective):
    def __init__(self):
        super().__init__()
        self.grad_fn = MeanAbsolutebackward

    def acc(self, y_pred: Variable, y_true: Variable):
        return 0.

    def forward(self, y_pred: Variable, y_true: Variable):
        loss_val = GlobalGraph.np.sum(GlobalGraph.np.absolute(y_pred.data - y_true.data)) / y_pred.data.shape[0]
        loss = Variable(data=loss_val, in_bounds=[y_pred, y_true])
        loss.grad_fn = self.grad_fn
        return loss


class BinaryCrossEntropy(Objective):
    def __init__(self):
        super().__init__()
        self.grad_fn = BinaryCrossEntropyBackward

    def acc(self, y_pred: Variable, y_true: Variable):
        pred = y_pred.data >= 0.5
        return GlobalGraph.np.mean(pred == y_true.data)

    def forward(self, y_pred: Variable, y_true: Variable):
        loss_val = -GlobalGraph.np.multiply(y_true.data, GlobalGraph.np.log(y_pred.data)) - GlobalGraph.np.multiply(1 - y_true.data, GlobalGraph.np.log(1 - y_pred.data))
        loss_val = GlobalGraph.np.sum(loss_val) / y_pred.data.shape[0]
        loss = Variable(data=loss_val, in_bounds=[y_pred, y_true])
        loss.grad_fn = self.grad_fn
        return loss


class SparseCrossEntropy(Objective):
    # used for one-hot label
    def __init__(self):
        super().__init__()
        self.grad_fn = SparseCrossEntropyBackward

    def __call__(self, y_pred: Variable, y_true: Variable):
        return super().__call__(y_pred, y_true)

    def acc(self, y_pred: Variable, y_true: Variable):
        acc = GlobalGraph.np.argmax(y_pred.data, axis=-1) == GlobalGraph.np.argmax(y_true.data, axis=-1)
        return GlobalGraph.np.mean(acc).tolist()

    def forward(self, y_pred: Variable, y_true: Variable):
        y_pred = softmax(y_pred)
        avg = GlobalGraph.np.prod(GlobalGraph.np.asarray(y_pred.shape[:-1]))
        loss_val = -GlobalGraph.np.sum(GlobalGraph.np.multiply(y_true.data, GlobalGraph.np.log(y_pred.data))) / avg
        loss = Variable(data=loss_val, in_bounds=[y_pred, y_true])
        loss.grad_fn = self.grad_fn
        return loss


class CrossEntropy(Objective):
    def __init__(self):
        super().__init__()
        self.grad_fn = CrossEntropyBackward

    def __call__(self, y_pred: Variable, y_true: Variable):
        return super().__call__(y_pred, y_true)

    def acc(self, y_pred: Variable, y_true: Variable):
        acc = (GlobalGraph.np.argmax(y_pred.data, axis=-1) == y_true.data)
        return GlobalGraph.np.mean(acc).tolist()

    def forward(self, y_pred: Variable, y_true: Variable):
        y_true.data = y_true.data.astype(GlobalGraph.np.int64)
        y_pred = softmax(y_pred)
        to_sum_dim = reduce(lambda x, y: x * y, y_pred.data.shape[:-1])
        last_dim = y_pred.data.shape[-1]
        n = y_pred.data.shape[0]
        probs = y_pred.data.reshape(-1, last_dim)
        y_flat = y_true.data.reshape(to_sum_dim)
        loss_val = - GlobalGraph.np.sum(GlobalGraph.np.log(probs[GlobalGraph.np.arange(to_sum_dim), y_flat])) / n
        loss = Variable(data=loss_val, in_bounds=[y_pred, y_true])
        loss.grad_fn = self.grad_fn
        return loss


def get_objective(objective):
    if objective.__class__.__name__ == 'str':
        objective = objective.lower()
        if objective in['crossentropy', 'cross_entropy']:
            return CrossEntropy()
        elif objective in['sparsecrossentropy', 'sparse_crossentropy', 'sparse_cross_entropy']:
            return SparseCrossEntropy()
        elif objective in ['binarycrossentropy', 'binary_cross_entropy', 'binary_crossentropy']:
            return BinaryCrossEntropy()
        elif objective in ['meansquarederror', 'mean_squared_error', 'mse']:
            return MeanSquaredError()
        elif objective in ['meanabsoluteerror', 'mean_absolute_error', 'mae']:
            return MeanAbsoluteError()
    elif isinstance(objective, Objective):
        return copy.deepcopy(objective)
    else:
        raise ValueError('unknown objective type!')



from ..nn.core import Layer, Variable
from xshinnosuke.layers.activators import get_activator
from ..utils.initializers import get_initializer
from ..nn.functional import concatenate
from ..utils.toolkit import initialize_ops_grad
from ..nn.global_graph import np
from typing import Union, List, Tuple


class Cell:
    def __init__(self, units: int, activation: str = 'tanh', initializer: str = 'glorotuniform',
                 recurrent_initializer: str = 'orthogonal', **kwargs):
        self.units = units
        # 不知道c++支持这样吗，比如有个变量a是由Variable这个类实例化的， 那么a.__class__就返回的是a的类，既Variable，然后使用self.activator_cls()和Variable()等价
        self.activation_cls = get_activator(activation).__class__
        self.initializer = get_initializer(initializer)
        self.recurrent_initializer = get_initializer(recurrent_initializer)
        self.variables = []  # Variable数组
        self.activations = []  # Activation这个类的数组

    # 以下全都由子类实现
    def initial_params(self, input_shape=None):
        pass

    def forward(self, x: Variable, is_training: bool, stateful: bool):
        pass

    def backward(self, outputs: Variable, inputs: Variable, return_sequences: bool):
        pass

    def reset_state(self, shape):
        pass


class Recurrent(Layer):
    # Base class for Recurrent layer
    def __init__(self, cell: Cell, return_sequences: bool = False, return_state: bool = False, stateful: bool = False,
                 input_length: int = None, **kwargs):
        '''
        :param cell: A Cell object
        :param return_sequences: return all output sequences if true,else return output sequences' last output
        :param return_state:if true ,return last state
        :param stateful:if true，the sequences last state will be used as next sequences initial state
        :param input_length: input sequences' length
        :param kwargs:
        '''
        super(Recurrent, self).__init__(**kwargs)
        self.cell = cell
        self.return_sequences = return_sequences
        self.return_state = return_state
        self.stateful = stateful
        self.input_length = input_length

    def __call__(self, inbound):
        assert len(inbound.shape) == 2, 'Only support batch input'
        super(Recurrent, self).__call__(inbound)
        return self

    def compute_output_shape(self, input_shape: Union[List, Tuple] = None) -> Union[List, Tuple]:
        time_steps, n_vec = input_shape
        if self.return_sequences:
            return tuple([time_steps, self.cell.units])
        return tuple([self.cell.units])

    def initial_params(self, *args):
        time_steps, n_vec = self.input_shape
        self.variables = self.cell.initial_params((n_vec, self.cell.units))

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x
        output = self.cell.forward(self.input_data, is_training, self.stateful)
        if self.return_sequences:
            self.data = output
        else:
            self.data = output[:, -1, :]

        self.connect_init(self.data, is_training)
        if self.return_sequences:
            return self.data, self.cell.prev_a
        return self.data

    def backward(self, gradients=None):
        gradients = self.cell.backward(self.data, self.input_data, self.return_sequences)
        super().backward(gradients)


class SimpleRNNCell(Cell):
    # cell class for SimpleRNN
    def __init__(self, units: int, activation: str = 'tanh', initializer: str = 'glorotuniform',
                 recurrent_initializer: str = 'orthogonal',  **kwargs):
        super(SimpleRNNCell, self).__init__(units=units, activation=activation, initializer=initializer,
                                            recurrent_initializer=recurrent_initializer)

        self.__first_initialize = True  # bool
        self.prev_a = None  # ndarray
        self.time_steps = None  # int

    def initial_params(self, input_shape=None):
        n_in, n_out = input_shape
        Wxa = Variable(self.initializer((n_in, n_out)), name='variable')
        Waa = Variable(self.recurrent_initializer((n_out, n_out)), name='variable')
        ba = Variable(np.zeros(n_out), name='variable')
        self.variables.append(Wxa)
        self.variables.append(Waa)
        self.variables.append(ba)
        return self.variables

    def forward(self, x: Variable, is_training: bool, stateful: bool):
        batch_nums, time_steps, n_vec = x.data.shape
        Wxa, Waa, ba = self.variables
        if is_training:
            self.time_steps = time_steps
            if self.__first_initialize:
                # first initialize prev_a
                self.reset_state(shape=(batch_nums, time_steps + 1, self.units))
                self.activations = [self.activation_cls() for _ in range(time_steps)]
                self.__first_initialize = False

            if stateful:
                self.prev_a[:, 0, :] = self.prev_a[:, -1, :]
            else:
                self.reset_state(shape=(batch_nums, time_steps + 1, self.units))
        else:
            self.reset_state(shape=(batch_nums, time_steps + 1, self.units))

        for t in range(1, time_steps + 1):
            self.prev_a[:, t, :] = self.activations[t-1].forward(x.data[:, t-1, :].dot(Wxa.data) +
                                                                 self.prev_a[:, t-1, :].dot(Waa.data) + ba.data)

        initialize_ops_grad(self.variables)
        return self.prev_a[:, 1:, :]

    def backward(self, outputs: Variable, inputs: Variable, return_sequences: bool):
        Wxa, Waa, ba = self.variables
        grad = np.zeros_like(inputs.data)
        if return_sequences:
            da_next = np.zeros_like(self.prev_a[:, 0, :])
            for t in reversed(range(self.time_steps)):
                dz = self.activations[t].backward(outputs.grad[:, t, :] + da_next)

                if Waa.requires_grad:
                    Waa.grad += np.dot(self.prev_a[:, t - 1, :].T, dz)
                if Wxa.requires_grad:
                    Wxa.grad += np.dot(inputs.data[:, t, :].T, dz)
                if ba.requires_grad:
                    ba.grad += np.sum(dz, axis=0)

                da_next = np.dot(dz, Waa.data.T)
                grad[:, t, :] = np.dot(dz, Wxa.data.T)
        else:
            da = outputs.grad
            for t in reversed(range(self.time_steps)):
                da = self.activations[t].backward(da)

                if Waa.requires_grad:
                    Waa.grad += np.dot(self.prev_a[:, t-1, :].T, da)
                if Wxa.requires_grad:
                    Wxa.grad += np.dot(inputs.data[:, t, :].T, da)
                if ba.requires_grad:
                    ba.grad += np.sum(da, axis=0)

                grad[:, t, :] = np.dot(da, Wxa.data.T)
                da = np.dot(da, Waa.data.T)
        return grad

    def reset_state(self, shape):
        self.prev_a = np.zeros(shape)


class SimpleRNN(Recurrent):
    # Fully-connected RNN
    def __init__(self, units: int, activation: str = 'tanh', initializer: str = 'glorotuniform',
                 recurrent_initializer: str = 'orthogonal', return_sequences: bool = False, return_state: bool = False,
                 stateful: bool = False, **kwargs):
        cell = SimpleRNNCell(units=units, activation=activation, initializer=initializer,
                             recurrent_initializer=recurrent_initializer)
        super(SimpleRNN, self).__init__(cell, return_sequences=return_sequences, return_state=return_state,
                                        stateful=stateful, **kwargs)


class LSTMCell(Cell):
    # cell class for SimpleRNN
    def __init__(self, units: int, activation: str = 'tanh', recurrent_activation: str = 'sigmoid',
                 initializer: str = 'glorotuniform', recurrent_initializer: str = 'orthogonal', unit_forget_bias: bool = True, **kwargs):
        super(LSTMCell, self).__init__(units=units, activation=activation, initializer=initializer,
                                       recurrent_initializer=recurrent_initializer)

        self.recurrent_activation_cls = get_activator(recurrent_activation).__class__
        self.unit_forget_bias = unit_forget_bias
        self.recurrent_activations = []  # Activation这个类的数组
        self.activations = []  # Activation这个类的数组
        self.__first_initialize = True  # bool
        self.time_steps = None  # int
        # 以下都是ndarray，具体看reset_state()
        self.prev_a = None
        self.c = None
        self.tao_f = None
        self.tao_u = None
        self.tao_o = None
        self.c_tilde = None


    def initial_params(self, input_shape=None):
        n_in, n_out = input_shape
        # Wf_l means forget gate linear weight,Wf_r represents forget gate recurrent weight.

        # forget gate
        Wf_l = Variable(self.initializer((n_in, n_out)), name='variable')
        Wf_r = Variable(self.recurrent_initializer((n_out, n_out)), name='variable')
        # update gate
        Wu_l = Variable(self.initializer((n_in, n_out)), name='variable')
        Wu_r = Variable(self.recurrent_initializer((n_out, n_out)), name='variable')
        # update unit
        Wc_l = Variable(self.initializer((n_in, n_out)), name='variable')
        Wc_r = Variable(self.recurrent_initializer((n_out, n_out)), name='variable')
        # output gate
        Wo_l = Variable(self.initializer((n_in, n_out)), name='variable')
        Wo_r = Variable(self.recurrent_initializer((n_out, n_out)), name='variable')

        Wf = concatenate(Wf_r, Wf_l, axis=0, name='variable')
        Wu = concatenate(Wu_r, Wu_l, axis=0, name='variable')
        Wc = concatenate(Wc_r, Wc_l, axis=0, name='variable')
        Wo = concatenate(Wo_r, Wo_l, axis=0, name='variable')
        W = concatenate(Wf, Wu, Wc, Wo, axis=1)

        if self.unit_forget_bias:
            bf = Variable(np.ones((1, n_out)), name='variable')
        else:
            bf = Variable(np.zeros((1, n_out)), name='variable')
        bu = Variable(np.zeros((1, n_out)), name='variable')
        bc = Variable(np.zeros((1, n_out)), name='variable')
        bo = Variable(np.zeros((1, n_out)), name='variable')
        b = concatenate(bf, bu, bc, bo, axis=1, name='variable')
        del Wf_r, Wf_l, Wu_r, Wu_l, Wc_r, Wc_l, Wo_r, Wo_l, Wf, Wu, Wc, Wo, bf, bu, bc, bo
        self.variables.append(W)
        self.variables.append(b)
        return self.variables

    def forward(self, x: Variable, is_training: bool, stateful: bool):
        batch_nums, time_steps, n_vec = x.data.shape
        W, b = self.variables
        if is_training:
            self.time_steps = time_steps
            if self.__first_initialize:
                # first intialize prev_a
                self.reset_state(shape=(batch_nums, time_steps + 1, self.units))
                self.activations = [self.activation_cls() for _ in range(time_steps)]
                self.recurrent_activations = [self.recurrent_activation_cls() for _ in range(3 * time_steps)]
                self.__first_initialize = False
            if stateful:
                self.prev_a[:, 0, :] = self.prev_a[:, -1, :]
                self.c[:, 0, :] = self.c[:, -1, :]
            else:
                self.reset_state(shape=(batch_nums, time_steps + 1, self.units))
        else:
            self.reset_state(shape=(batch_nums, time_steps + 1, self.units))

        z = np.zeros((batch_nums, time_steps, n_vec + self.units))
        for t in range(1, time_steps + 1):
            zt = np.concatenate((self.prev_a[:, t-1, :], x.data[:, t-1, :]), axis=1)
            ot = zt.dot(W.data) + b.data
            f = self.recurrent_activations[3 * (t-1)].forward(ot[:, :self.units])
            u = self.recurrent_activations[3 * (t - 1) + 1].forward(ot[:, self.units: self.units * 2])
            c_tilde = self.activations[t - 1].forward(ot[:, self.units * 2: self.units * 3])
            o = self.recurrent_activations[3 * (t - 1) + 2].forward(ot[:, self.units * 3:])
            self.c_tilde[:, t-1, :] = c_tilde
            c = f * self.c[:, t-1, :] + u * c_tilde
            self.prev_a[:, t, :] = o * np.tanh(c)

            self.tao_f[:, t-1, :] = f
            self.tao_u[:, t-1, :] = u
            self.tao_o[:, t - 1, :] = o
            self.c[:, t, :] = c
            z[:, t-1, :] = zt
        return self.prev_a[:, 1:, :]

    def backward(self, outputs: Variable, inputs: Variable, return_sequences: bool):
        W, b = self.variables
        grad = np.zeros_like(inputs.data)
        grad = grad[:, :, self.units:]
        if return_sequences:
            da_next = np.zeros_like(self.prev_a[:, 0, :])
            dc_next = np.zeros_like(self.c[:, 0, :])
            for t in reversed(range(self.time_steps)):
                da = outputs.grad[:, t, :] + da_next
                dtao_o = da * np.tanh(self.c[:, t + 1, :])
                do = self.recurrent_activations[3 * (t + 1) - 1].backward(dtao_o)
                dc = dc_next
                dc += da * self.tao_o[:, t, :] * (1 - np.square(np.tanh(self.c[:, t+1, :])))
                dc_tilde = dc * self.tao_u[:, t, :]
                dc_tilde_before_act = self.activations[t].backward(dc_tilde)
                dtao_u = dc*self.c_tilde[:, t, :]
                du = self.recurrent_activations[3 * (t + 1) - 2].backward(dtao_u)
                dtao_f = dc * self.c[:, t, :]
                df = self.recurrent_activations[3 * (t + 1) - 3].backward(dtao_f)
                dgrad = np.concatenate((do, dc_tilde_before_act, du, df), axis=1)
                if W.requires_grad:
                    W.grad += np.dot(inputs.data[:, t, :].T, dgrad)
                if b.requires_grad:
                    b.grad += np.sum(dgrad, axis=0, keepdims=True)

                dz = dgrad.dot(W.data.T)

                da_next = dz[:, :self.units]
                dc_next = dc * self.tao_f[:, t, :]

                grad[:, t, :] = dz[:, self.units:]
        else:
            da_next = np.zeros_like(self.prev_a[:, 0, :])
            dc_next = np.zeros_like(self.c[:, 0, :])
            da = outputs.grad + da_next
            for t in reversed(range(self.time_steps)):
                dtao_o = da * np.tanh(self.c[:, t+1, :])
                do = self.recurrent_activations[3 * (t + 1) - 1].backward(dtao_o)

                dc = dc_next
                dc += da * self.tao_o[:, t, :] * (1 - np.square(np.tanh(self.c[:, t+1, :])))

                dc_tilde = dc * self.tao_u[:, t, :]
                dc_tilde_before_act = self.activations[t].backward(dc_tilde)

                dtao_u = dc * self.c_tilde[:, t, :]
                du = self.recurrent_activations[3 * (t + 1) - 2].backward(dtao_u)

                dtao_f = dc * self.c[:, t, :]
                df = self.recurrent_activations[3 * (t + 1) - 3].backward(dtao_f)

                dgrad = np.concatenate((do, dc_tilde_before_act, du, df), axis=1)
                if W.requires_grad:
                    W.grad += np.dot(inputs.data[:, t, :].T, dgrad)
                if b.requires_grad:
                    b.grad += np.sum(dgrad, axis=0, keepdims=True)

                dz = dgrad.dot(W.data.T)

                da = dz[:, :self.units]
                dc_next = dc * self.tao_f[:, t, :]
                grad[:, t, :] = dz[:, self.units:]

        return grad

    def reset_state(self, shape):
        # timesteps here equals to real timesteps+1
        batch_nums, time_steps, units = shape
        self.prev_a  = np.zeros(shape)
        self.c = np.zeros(shape)
        self.tao_f = np.zeros((batch_nums, time_steps - 1, units))
        self.tao_u = np.zeros((batch_nums, time_steps - 1, units))
        self.tao_o = np.zeros((batch_nums, time_steps - 1, units))
        self.c_tilde = np.zeros((batch_nums, time_steps - 1, units))


class LSTM(Recurrent):
    # Fully-connected RNN
    def __init__(self, units: int, activation: str = 'tanh', recurrent_activation: str = 'sigmoid', initializer: str = 'glorotuniform', recurrent_initializer: str = 'orthogonal', unit_forget_bias: bool = True, return_sequences: bool = False, return_state: bool = False, stateful: bool = False, **kwargs):
        '''
        :param units: hidden unit nums
        :param activation:  update unit activation
        :param recurrent_activation: forget gate,update gate,and output gate activation
        :param initializer: same to activation
        :param recurrent_initializer:same to recurrent_activation
        :param unit_forget_bias:if True,add one to the forget gate bias,and force bias initialize as zeros
        :param return_sequences: return sequences or last output
        :param return_state: if True ,return output and last state
        :param stateful: same as SimpleRNN
        :param kwargs:
        '''
        cell = LSTMCell(units=units, activation=activation, recurrent_activation=recurrent_activation, initializer=initializer, recurrent_initializer=recurrent_initializer, unit_forget_bias=unit_forget_bias)
        super(LSTM, self).__init__(cell, return_sequences=return_sequences, return_state=return_state,
                                   stateful=stateful, **kwargs)



from ..utils.initializers import get_initializer
from xshinnosuke.layers.activators import get_activator
from ..nn.core import Layer, Variable
from ..nn import global_graph as GlobalGraph
from ..nn.functional import conv2d, max_pool2d, avg_pool2d
from ..nn.grad_fn import Conv2DBackward, Maxpool2DBackward, Avgpool2DBackward
from typing import Tuple, Union


class Conv2D(Layer):
    def __init__(self, out_channels: int, kernel_size: Union[int, Tuple], use_bias: bool = False, stride: Union[int, Tuple] = 1, padding: Union[int, str] = 0,
                 activation: str = None, kernel_initializer: str = 'Normal', bias_initializer: str = 'zeros',
                 input_shape: Tuple = None,
                 **kwargs):
        self.out_channels = out_channels
        self.kernel_size = self.__check_size(kernel_size)
        self.input_shape = input_shape
        self.use_bias = use_bias
        self.stride = self.__check_size(stride)
        self.padding = padding
        self.activation = get_activator(activation) if activation is not None else None
        self.kernel_initializer = get_initializer(kernel_initializer)
        self.bias_initializer = get_initializer(bias_initializer)
        self.pad_size = padding
        self.cols = None
        super(Conv2D, self).__init__(input_shape=input_shape, **kwargs)

    def __check_size(self, kernel_size):
        if isinstance(kernel_size, (int, float)):
            return int(kernel_size), int(kernel_size)
        return kernel_size

    def initial_params(self, input_shape: Tuple = None):
        if input_shape is not None:
            self.input_shape = input_shape
        w = Variable(self.kernel_initializer((self.out_channels, self.input_shape[0], self.kernel_size[0],
                                              self.kernel_size[1])), name='variable')
        if self.use_bias:
            b = Variable(self.bias_initializer(1, self.out_channels), name='variable')
        else:
            b = None
        self.variables.append(w)
        self.variables.append(b)

    def compute_output_shape(self, input_shape: Tuple = None):
        assert len(input_shape) == 3
        n_C_prev, n_H_prev, n_W_prev = input_shape
        filter_h, filter_w = self.kernel_size
        if self.padding.__class__.__name__ == 'str':
            padding = self.padding.upper()
            if padding == 'SAME':
                n_H = n_H_prev
                n_W = n_W_prev
                pad_h = (self.stride[0] * (n_H_prev - 1) - n_H_prev + filter_h) // 2
                self.pad_size = pad_h
            elif padding == 'VALID':
                n_H = (n_H_prev - filter_h) // self.stride[0] + 1
                n_W = (n_W_prev - filter_w) // self.stride[1] + 1
                self.pad_size = 0
            else:
                raise TypeError('Unknown padding type!plz inputs SAME or VALID or an integer')
        else:
            assert isinstance(self.padding, int)
            n_H = (n_H_prev - filter_h + 2 * self.padding) // self.stride[0] + 1
            n_W = (n_W_prev - filter_w + 2 * self.padding) // self.stride[1] + 1
            self.pad_size = self.padding

        return self.out_channels, n_H, n_W

    def __call__(self, inbound):
        # if isinstance(inbound, Variable):
        if inbound.data is not None:
            if GlobalGraph.inputs is None:
                GlobalGraph.inputs = inbound

            if len(self.variables) == 0:
                self.initial_params(inbound.shape[1:])
            output = conv2d(inbound, self.variables[0], self.variables[1], self.stride, self.pad_size)
            if self.activation is not None:
                output = self.activation.__call__(output)
            # output是一个Variable
            return output

        super(Conv2D, self).__call__(inbound)
        return self

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x
        w, b = self.variables
        self.data = conv2d(self.input_data, w, b, self.stride, self.pad_size)
        if self.activation is not None:
            output = self.activation.forward(self.data)
            self.connect_init(output, is_training)
            return output
        else:
            self.connect_init(self.data, is_training)
            return self.data

    def backward(self, gradients=None):
        if self.activation is not None:
            self.activation.backward()
        Conv2DBackward(self.data)


class MaxPooling2D(Layer):
    def __init__(self, kernel_size: int = 2, stride: int = None, padding: int = 0):
        self.kernel_size = kernel_size
        self.stride = self.__check_size(kernel_size) if stride is None else self.__check_size(stride)
        self.padding = padding
        self.mode = 'reshape'
        super(MaxPooling2D, self).__init__()

    def compute_output_shape(self, input_shape=None):
        n_c, n_h_prev, n_w_prev = input_shape
        if self.kernel_size == self.stride:
            self.mode = 'reshape'
        else:
            self.mode = 'im2col'
        n_h, n_w = (n_h_prev - self.kernel_size + 2 * self.padding) // self.stride[0] + 1, (n_w_prev - self.kernel_size
                                                                                            + 2 * self.padding) // self.stride[1] + 1
        return n_c, n_h, n_w

    def __check_size(self, kernel_size):
        if isinstance(kernel_size, (int, float)):
            return int(kernel_size), int(kernel_size)
        return kernel_size

    def __call__(self, inbound):
        if isinstance(inbound, Variable):
            output = max_pool2d(inbound, self.kernel_size, self.stride, self.padding)
            # output是一个Variable
            return output
        super(MaxPooling2D, self).__call__(inbound)
        return self

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x
        self.data = max_pool2d(self.input_data, self.kernel_size, self.stride, self.padding)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        Maxpool2DBackward(self.data)


class AvgPooling2D(Layer):
    def __init__(self, kernel_size: int = 2, stride: int = None, padding: int = 0):
        self.kernel_size = kernel_size
        self.stride = self.__check_size(kernel_size) if stride is None else self.__check_size(stride)
        self.padding = padding
        super(AvgPooling2D, self).__init__()

    def __check_size(self, kernel_size):
        if isinstance(kernel_size, (int, float)):
            return int(kernel_size), int(kernel_size)
        return kernel_size

    def compute_output_shape(self, input_shape=None):
        n_c, n_h_prev, n_w_prev = input_shape
        n_h, n_w = (n_h_prev - self.kernel_size + 2 * self.padding) // self.stride[0] + 1, \
                   (n_w_prev - self.kernel_size + 2 * self.padding) // self.stride[1] + 1
        return n_c, n_h, n_w

    def __call__(self, inbound):
        if isinstance(inbound, Variable):
            output = avg_pool2d(inbound, self.kernel_size, self.stride, self.padding)
            # output是一个Variable
            return output
        super(AvgPooling2D, self).__call__(inbound)
        return self

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x
        self.data = avg_pool2d(self.input_data, self.kernel_size, self.stride, self.padding)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        Avgpool2DBackward(self.data)



from ..nn.core import Layer, Variable
from ..nn.global_graph import np
from xshinnosuke.layers.activators import get_activator
from ..nn.functional import pad_2d
from ..nn.grad_fn import Pad2DBackward, NegBackward, MultiplyBackward, MatmulBackward, LogBackward, ExpBackward, SumBackward, MeanBackward, AbsBackward, PowBackward
from typing import Tuple, Union, List


class Input(Layer):
    def __init__(self, input_shape: Union[List, Tuple], **kwargs):
        super(Input, self).__init__(input_shape=input_shape, **kwargs)
        self.input_shape = input_shape
        self.shape = self.compute_output_shape(input_shape)

    def __call__(self, inbound):
        if isinstance(inbound, Variable):
            return inbound
        super().__call__(inbound)

    def compute_output_shape(self, input_shape: Union[List, Tuple] = None) -> Union[List, Tuple]:
        if input_shape is None:
            return self.input_shape
        return input_shape

    def forward(self, x: Variable = None, is_training:bool = True, *args):
        if x is not None:
            self.input_data = x
        self.data = self.input_data
        self.connect_init(self.data, is_training)
        return self.data


class ZeroPadding2D(Layer):
    def __init__(self, pad_size: Union[int, float, Tuple], **kwargs):
        self.pad_h, self.pad_w = self.__check_pad_size(pad_size)
        super(ZeroPadding2D, self).__init__(**kwargs)

    def __check_pad_size(self, pad_size: Union[int, float, Tuple]):
        if isinstance(pad_size, (int, float)):
            return int(pad_size), int(pad_size)
        return pad_size

    def __call__(self, inbound):
        if isinstance(inbound, Variable):
            output = pad_2d(inbound, (self.pad_h, self.pad_w))
            # output是一个Variable
            return output
        super(ZeroPadding2D, self).__call__(inbound)
        return self

    def compute_output_shape(self, input_shape: Union[List, Tuple] = None) ->  Union[List, Tuple]:
        c, h, w = input_shape
        return c, h + 2 * self.pad_h, w + 2 * self.pad_w

    def forward(self, x: Variable = None, is_training: bool = True, *args):
        if x is not None:
            self.input_data = x
        self.data = pad_2d(self.input_data, (self.pad_h, self.pad_w))
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        Pad2DBackward(self.data)


class Add(Layer):
    def __call__(self, inbounds: List[Layer]):
        # inbounds只能是Layer数组
        for inbound in inbounds:
            inbound.out_bounds.append(self)
            self.in_bounds.append(inbound)
            self.shape = inbound.shape
        return self

    def forward(self, x: Variable = None, is_training: bool = True, *args):
        if x is not None:
            self.input_data = x
        data = 0
        for in_bound in self.in_bounds:
            data += in_bound.data.data
        self.data = Variable(data, in_bounds=self.in_bounds)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        for in_bound in self.in_bounds:
            if in_bound.data.requires_grad:
                in_bound.data.grad += self.data.grad


class Negative(Layer):
    def __call__(self, inbound: Layer):
        super().__call__(inbound)
        return self

    def forward(self, x: Variable = None, is_training: bool = True, *args):
        if x is not None:
            self.input_data = x
        self.data = Variable(data=-self.input_data.data, in_bounds=self.in_bounds)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        NegBackward(self.data)


class Multiply(Layer):
    def __call__(self, inbounds: List[Layer]):
        for inbound in inbounds:
            super().__call__(inbound)
        return self

    def forward(self, x: Variable = None, is_training: bool = True, *args):
        if x is not None:
            self.input_data = x
        data = 1
        for in_bound in self.in_bounds:
            data = np.multiply(data, in_bound.data)
        self.data = Variable(data=data, in_bounds=self.in_bounds)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        MultiplyBackward(self.data)


class Matmul(Layer):
    def __call__(self, inbounds: List[Layer]):
        """
        只支持两个Layer做矩阵乘
        """
        assert len(inbounds) == 2
        for inbound in inbounds:
            inbound.out_bounds.append(self)
            self.in_bounds.append(inbound)
        self.shape = inbounds[0].shape[:-1] + inbounds[-1].shape[1:]
        return self

    def forward(self, x: Variable = None, is_training: bool = True, *args):
        if x is not None:
            self.input_data = x
        data = self.in_bounds[0].data.dot(self.in_bounds[1].data)
        self.data = Variable(data=data, in_bounds=self.in_bounds)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        MatmulBackward(self.data)


class Log(Layer):
    def __init__(self, base: Union[int, str]='e'):
        self.base = self.__check_base(base)
        super().__init__()

    def __check_base(self, base: Union[int, str]):
        if base == 'e':
            return np.e
        elif base in [2, 10]:
            return base
        else:
            raise ValueError('unknown base value {}'.format(base))

    def __call__(self, inbound: Layer):
        super().__call__(inbound)
        return self

    def forward(self, x: Variable = None, is_training: bool = True, *args):
        if x is not None:
            self.input_data = x

        if self.base == 2:
            data = np.log2(self.in_bounds[0].data)
        elif self.base == 10:
            data = np.log10(self.in_bounds[0].data)
        else:
            data = np.log(self.in_bounds[0].data)
        self.data = Variable(data=data, in_bounds=self.in_bounds)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        LogBackward(self.data)


class Exp(Layer):
    def __call__(self, inbound: Layer):
        super().__call__(inbound)
        return self

    def forward(self, x: Variable = None, is_training: bool = True, *args):
        if x is not None:
            self.input_data = x
        self.data = Variable(data=np.exp(self.in_bounds[0].data), in_bounds=self.in_bounds)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        ExpBackward(self.data)


class Sum(Layer):
    def __call__(self, inbound: Layer):
        super().__call__(inbound)
        return self

    def compute_output_shape(self, input_shape: Union[List, Tuple] = None) -> Union[List, Tuple]:
        return tuple([1])

    def forward(self, x: Variable = None, is_training: bool = True, *args):
        if x is not None:
            self.input_data = x
        self.data = Variable(data=np.sum(self.in_bounds[0].data), in_bounds=self.in_bounds)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        SumBackward(self.data)


class Mean(Layer):
    def __call__(self, inbound: Layer):
        super().__call__(inbound)
        return self

    def compute_output_shape(self, input_shape: Union[List, Tuple] = None) -> Union[List, Tuple]:
        return tuple([1])

    def forward(self, x: Variable = None, is_training: bool = True, *args):
        if x is not None:
            self.input_data = x
        self.data = Variable(data=np.mean(self.in_bounds[0].data), in_bounds=self.in_bounds)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        MeanBackward(self.data)


class Abs(Layer):
    def __call__(self, inbound: Layer):
        super().__call__(inbound)
        return self

    def forward(self, x: Variable = None, is_training: bool = True, *args):
        if x is not None:
            self.input_data = x
        self.data = Variable(data=np.abs(self.in_bounds[0].data), in_bounds=self.in_bounds)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        AbsBackward(self.data)


class Pow(Layer):
    def __init__(self, exponent:int = 2):
        self.exponent = exponent
        super().__init__()

    def __call__(self, inbound: Layer):
        super().__call__(inbound)
        return self

    def forward(self, x: Variable = None, is_training: bool = True, *args):
        if x is not None:
            self.input_data = x
        self.data = Variable(data=np.power(self.in_bounds[0].data, self.exponent), in_bounds=self.in_bounds)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        PowBackward(self.data)



from .base import *
from .convolutional import *
from .embeddings import *
from .linear import *
from .normalize import *
from .recurrent import *
from .wrappers import *


name = 'xshinnosuke-layers'



from ..nn.core import Layer, Variable
from ..nn.global_graph import np
from typing import Union, Tuple, List


class TimeDistributed(Layer):
    # 该层主要用于对时间序列数据做操作的，比如对时间序列使用全连接或者卷积
    def __init__(self, layer: Layer, **kwargs):
        super(TimeDistributed, self).__init__(**kwargs)
        # 所有时间步上的layer是共享的
        self.layer = layer

    def initial_params(self, *args):
        self.layer.initial_params(self.input_shape[1:])
        self.variables = self.layer.variables

    def compute_output_shape(self, input_shape: Union[List, Tuple] = None) -> Union[List, Tuple]:
        time_steps = input_shape[0]
        out_shape = self.layer.compute_output_shape(input_shape[1:])
        return (time_steps, ) + out_shape

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x
        time_steps = self.input_shape[0]
        output = Variable(data=np.zeros((self.input_data.data.shape[0], ) + self.shape))
        for t in range(time_steps):
            output[:, t] = self.layer.forward(self.input_data[:, t])
        self.data = output
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        time_steps = self.input_shape[0]
        for t in range(time_steps):
            self.data.in_bounds[t].grad = self.data.grad[:, t]
            self.layer.data = self.data.in_bounds[t]
            self.layer.backward()
            self.input_data.grad[:, t] = self.layer.data.in_bounds[0].grad
        super().backward(self.input_data.grad)



from xshinnosuke.nn.core import Layer, Variable
from xshinnosuke.nn.grad_fn import ReluBackward, SigmoidBackward, TanhBackward
from xshinnosuke.nn.functional import relu, sigmoid, tanh


class Activation(Layer):
    def __init__(self, act_name: str = 'relu'):
        self.activation = get_activator(act_name)
        super(Activation, self).__init__()

    def __call__(self, inbound):
        if isinstance(inbound, Variable):
            output = self.activation.forward(inbound)
            # output是一个Variable
            return output
        super(Activation, self).__call__(inbound)
        return self

    def compute_output_shape(self, input_shape=None):
        return input_shape

    def forward(self, x: Variable = None, is_training: bool = True, *args):
        if x is not None:
            self.input_data = x
        self.data = self.activation.forward(self.input_data)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        self.activation.backward()


class ReLU(Layer):
    def __init__(self, inplace=False):
        super().__init__()
        self.inplace = inplace

    def __call__(self, inbound):
        if isinstance(inbound, Variable):
            # output = relu(inbound, self.inplace)
            #
            # return output
            return self.forward(inbound)
        super().__call__(inbound)
        return self

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x
        self.data = relu(self.input_data, inplace=self.inplace)
        self.data.cache['inplace'] = self.inplace
        if self.inplace:
            if 'grad_fn' not in self.data.cache:
                self.data.cache['grad_fn'] = []
            self.data.cache['grad_fn'].append(self.data.grad_fn)
            self.data.cache['mask'] = self.input_data.data < 0
            self.data.grad_fn = ReluBackward
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, output: Variable = None):
        if self.inplace:
            raise RuntimeError('a leaf Variable that requires grad is being used in an in-place operation.')
        if output is not None:
            self.data = output
        ReluBackward(self.data)


class Linear(Layer):
    def __call__(self, inbound):
        if isinstance(inbound, Variable):
            return inbound
        super().__call__(inbound)
        return self

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x
        self.data = self.input_data
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, output: Variable = None):
        if output is not None:
            self.data = output
        self.input_data.grad = self.data.grad
        return self.input_data


class Sigmoid(Layer):
    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x
        self.data = sigmoid(self.input_data)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, output: Variable = None):
        if output is not None:
            self.data = output
        SigmoidBackward(self.data)


class Tanh(Layer):
    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x
        self.data = tanh(self.input_data)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, output: Variable = None):
        if output is not None:
            self.data = output
        TanhBackward(self.data)


def get_activator(activator):
    if activator.__class__.__name__ == 'str':
        activator = activator.lower()
        if activator == 'relu':
            return ReLU()
        elif activator == 'linear':
            return Linear()
        elif activator == 'sigmoid':
            return Sigmoid()
        elif activator == 'tanh':
            return Tanh()



# 写关于Node和Layer运算所需要的函数
from ..nn.global_graph import np
from ..utils.toolkit import col2im
from functools import reduce


def AddBackward(outputs):
    for in_bound in outputs.in_bounds:
        if in_bound.requires_grad:
            in_bound.grad += outputs.grad


def IAddBackward(outputs):
    inputs = outputs.in_bounds.pop()
    if inputs.requires_grad:
        inputs.grad += outputs.grad
    outputs.grad_fn = outputs.cache['grad_fn'].pop()
    outputs.grad_fn(outputs)


def NegBackward(outputs):
    inputs, = outputs.in_bounds
    if inputs.requires_grad:
        inputs.grad += -outputs.grad


def MatmulBackward(outputs):
    # z = x * y, dloss/dx = dloss/dz * dz/dx = dloss/dz * y
    x, y = outputs.in_bounds
    if x.requires_grad:
        x.grad += np.dot(outputs.grad, y.data.T)
    if y.requires_grad:
        y.grad += np.dot(x.data.T, outputs.grad)


def TransposeBackward(outputs):
    inputs, = outputs.in_bounds
    if inputs.requires_grad:
        inputs.grad += outputs.grad.reshape(inputs.shape)


def MultiplyBackward(outputs):
    length = len(outputs.in_bounds)
    product_except_self_list = [0] * length
    product_except_self_list[0] = 1
    for i in range(1, length):
        product_except_self_list[i] = product_except_self_list[i - 1] * outputs.in_bounds[i - 1].data
    right = 1
    for i in reversed(range(length)):
        product_except_self_list[i] *= right
        right *= outputs.in_bounds[i].data

    for i, in_bound in enumerate(outputs.in_bounds):
        if in_bound.requires_grad:
            in_bound.grad += outputs.grad * product_except_self_list[i]


def SumBackward(outputs):
    inputs, = outputs.in_bounds
    if inputs.requires_grad:
        inputs.grad += np.ones_like(inputs.data) * outputs.grad


def MeanBackward(outputs):
    inputs, = outputs.in_bounds
    if inputs.requires_grad:
        mean_nums = inputs.shape / outputs.shape
        inputs.grad += np.ones_like(inputs.data) * outputs.grad / mean_nums


def AbsBackward(outputs):
    inputs, = outputs.in_bounds
    if inputs.requires_grad:
        grad_ = np.asarray(outputs.data == inputs.data, dtype=np.int8)
        grad_[grad_ == 0] = -1
        inputs.grad += grad_


def ViewBackward(outputs):
    inputs, = outputs.in_bounds
    if inputs.requires_grad:
        inputs.grad += outputs.grad.reshape(inputs.shape)


def LogBackward(outputs):
    inputs, = outputs.in_bounds
    if inputs.requires_grad:
        base = outputs.cache['base'] if outputs.cache['base'] != 'e' else np.e
        inputs.grad += outputs.grad / (inputs.data * np.log(base))


def ExpBackward(outputs):
    inputs, = outputs.in_bounds
    if inputs.requires_grad:
        inputs.grad += outputs.grad * outputs.data


def PowBackward(outputs):
    inputs, = outputs.in_bounds
    if inputs.requires_grad:
        n_times = inputs.power
        if n_times < 1:
            inputs.grad += n_times * outputs.grad / np.power(inputs.data, 1 - n_times)
        else:
            inputs.grad += n_times * np.power(inputs.data, n_times - 1) * outputs.grad


def ReluBackward(outputs):
    if outputs.cache['inplace']:
        mask = outputs.cache['mask']
        outputs.grad[mask] = 0
        outputs.grad_fn = outputs.cache['grad_fn'].pop()
        outputs.grad_fn(outputs)
    else:
        inputs, = outputs.in_bounds
        if inputs.requires_grad:
            grad = outputs.grad
            grad[inputs.data < 0] = 0
            inputs.grad += grad


def SigmoidBackward(outputs):
    inputs, = outputs.in_bounds
    if inputs.requires_grad:
        inputs.grad += outputs.grad * outputs.data * (1 - outputs.data)


def TanhBackward(outputs):
    inputs, = outputs.in_bounds
    if inputs.requires_grad:
        inputs.grad += outputs.grad * (1 - np.square(outputs.data))


def FlattenBackward(outputs):
    inputs, = outputs.in_bounds
    if inputs.requires_grad:
        inputs.grad += outputs.grad.reshape(inputs.shape)


def DenseBackward(outputs):
    inputs, weight, bias = outputs.in_bounds
    # 正向传播时是 inputs -> dense -> outputs
    if weight.requires_grad:
        weight.grad += inputs.data.T.dot(outputs.grad)
    if bias is not None and bias.requires_grad:
        bias.grad += np.sum(outputs.grad, axis=0, keepdims=True)
    if inputs.requires_grad:
        inputs.grad += outputs.grad.dot(weight.data.T)


def Conv2DBackward(outputs):
    inputs, weight, bias = outputs.in_bounds
    n, in_channels, h, w = inputs.data.shape
    out_channels, _, kernel_h, kernel_w = weight.data.shape
    _, _, out_h, out_w = outputs.grad.shape
    grad_reshaped = outputs.grad.transpose(1, 0, 2, 3).reshape(out_channels, -1)
    if weight.requires_grad:
        weight.grad += grad_reshaped.dot(outputs.cache['col']).reshape(out_channels, in_channels, kernel_h, kernel_w)

    if bias is not None and bias.requires_grad:
        bias.grad += np.sum(outputs.grad, axis=(0, 2, 3))

    dcol = grad_reshaped.T.dot(weight.data.reshape(out_channels, -1))

    if inputs.requires_grad:
        inputs.grad += col2im(inputs.shape, outputs.cache['pad_size'], kernel_h, kernel_w,  outputs.cache['stride'],
                              dcol)


def Maxpool2DBackward(outputs):
    mode = outputs.cache['mode']
    inputs, = outputs.in_bounds
    if mode == 'reshape':
        dx_reshaped = np.zeros_like(outputs.cache['x_reshaped'])
        out_newaxis = outputs.data[:, :, :, np.newaxis, :, np.newaxis]
        mask = (outputs.cache['x_reshaped'] == out_newaxis)
        dout_newaxis = outputs.grad[:, :, :, np.newaxis, :, np.newaxis]
        dout_broadcast, _ = np.broadcast_arrays(dout_newaxis, dx_reshaped)
        dx_reshaped[mask] = dout_broadcast[mask]
        dx_reshaped /= np.sum(mask, axis=(3, 5), keepdims=True)
        grad = dx_reshaped.reshape(inputs.data.shape)
        if outputs.cache['pad_size']:
            grad = grad[:, :, outputs.cache['pad_size']: -outputs.cache['pad_size'], outputs.cache['pad_size']: -outputs.cache['pad_size']]
    else:
        grad = outputs.grad.transpose(0, 2, 3, 1)
        dmax = np.zeros((grad.size, outputs.cache['kernel_size'] * outputs.cache['kernel_size']))
        dmax[np.arange(outputs.cache['pool_argmax'].size), outputs.cache['pool_argmax'].flatten()] = grad.flatten()
        dmax = dmax.reshape(grad.shape + (outputs.cache['kernel_size'] * outputs.cache['kernel_size'],))

        dcol = dmax.reshape(reduce(lambda x, y: x * y, dmax.shape[:3]), -1)
        grad = col2im(inputs.shape, outputs.cache['pad_size'], outputs.cache['kernel_size'],
                         outputs.cache['kernel_size'], outputs.cache['stride'], dcol)

    if inputs.requires_grad:
        inputs.grad += grad


def Avgpool2DBackward(outputs):
    inputs, = outputs.in_bounds
    grad = outputs.grad.transpose(0, 2, 3, 1)
    dmean = np.repeat(grad.flatten(), outputs.cache['pool_argmean'].size) / (outputs.cache['kernel_size'] * outputs.cache['kernel_size'])
    dmean = dmean.reshape(grad.shape + (outputs.cache['kernel_size'] * outputs.cache['kernel_size'],))
    dcol = dmean.reshape(reduce(lambda x, y: x * y, dmean.shape[:3]), -1)
    grad = col2im(inputs.shape, outputs.cache['pad_size'], outputs.cache['kernel_size'],
                         outputs.cache['kernel_size'], outputs.cache['stride'], dcol)

    if inputs.requires_grad:
        inputs.grad += grad


def Pad2DBackward(outputs):
    inputs, = outputs.in_bounds
    pad_size = outputs.cache['pad_size']
    if inputs.requires_grad:
        inputs.grad += outputs.grad[:, :, pad_size[0]: -pad_size[0], pad_size[1]: -pad_size[1]]


def Dropout2DBackward(outputs):
    inputs, = outputs.in_bounds
    if inputs.requires_grad:
        inputs.grad += outputs.grad * outputs.cache['mask'] * outputs.cache['keep_prob']


def Batchnorm2DBackward(outputs):
    inputs, gamma, beta = outputs.in_bounds
    if inputs.requires_grad:
        grad = outputs.grad
        ndim = grad.ndim
        axis = inputs.cache['axis']
        xmu = inputs.cache['xmu']
        sqrtvar = inputs.cache['sqrtvar']
        normalized_x = inputs.cache['normalized_x']
        if not (axis == -1 or axis == ndim - 1):
            # for instance,inputs:(N,C,H,W),self.axis=1,after swapaxes,Inputs:(N,W,H,C)
            grad = np.swapaxes(grad, axis, -1)

        # (N,W,H,C) / (N,M)
        before_reshape_shape = grad.shape
        # (N*W*H,C) /(N,M)
        grad = grad.reshape(-1, inputs.data.shape[axis])

        if gamma.requires_grad:
            gamma.grad += np.sum(grad * normalized_x, axis=0)

        if beta.requires_grad:
            beta.grad += np.sum(grad, axis=0)

        N = normalized_x.shape[0]
        dnormalized_x = grad * gamma.data
        dvar = np.sum(np.power(- 1. / sqrtvar, 3) * xmu * dnormalized_x * 0.5, axis=0)
        dmean = np.sum(- dnormalized_x / sqrtvar, axis=0) - 2 * dvar * np.mean(xmu, axis=0)
        grad = dnormalized_x / sqrtvar + dvar * 2 * xmu / N + dmean / N
        grad = grad.reshape(before_reshape_shape)

        if not (axis == -1 or axis == ndim - 1):
            # for instance,outputs:(N,W,H,C),self.axis=1,after swapaxes,outputs:(N,C,H,W)
            grad = np.swapaxes(grad, axis, -1)

        inputs.grad += grad


def MeanSquaredbackward(outputs):
    y_pred, y_true = outputs.in_bounds
    gradients = outputs.grad * (y_pred.data - y_true.data) / y_pred.shape[0]
    if y_true.requires_grad:
        if y_true.grad is None:
            y_true.grad = gradients
        else:
            y_true.grad += gradients

    if y_pred.requires_grad:
        if y_pred.grad is None:
            y_pred.grad = gradients
        else:
            y_pred.grad += gradients


def MeanAbsolutebackward(outputs):
    y_pred, y_true = outputs.in_bounds
    pos = np.where((y_pred.data - y_true.data) < 0)
    mask = np.ones_like(y_pred.data)
    mask[pos] = -1

    gradients = mask / y_pred.data.shape[0]
    if y_true.requires_grad:
        if y_true.grad is None:
            y_true.grad = gradients
        else:
            y_true.grad += gradients

    if y_pred.requires_grad:
        if y_pred.grad is None:
            y_pred.grad = gradients
        else:
            y_pred.grad += gradients


def BinaryCrossEntropyBackward(outputs):
    y_pred, y_true = outputs.in_bounds
    avg = np.prod(np.asarray(y_pred.data.shape[:-1]))
    gradients = (np.divide(1 - y_true.data, 1 - y_pred.data) - np.divide(y_true.data, y_pred.data)) / avg
    if y_true.requires_grad:
        if y_true.grad is None:
            y_true.grad = gradients
        else:
            y_true.grad += gradients

    if y_pred.requires_grad:
        if y_pred.grad is None:
            y_pred.grad = gradients
        else:
            y_pred.grad += gradients


def SparseCrossEntropyBackward(outputs):
    y_pred, y_true = outputs.in_bounds
    # # before softmax
    before_softmax_y_pred = y_pred.in_bounds[0]
    avg = np.prod(np.asarray(y_pred.data.shape[:-1]))
    gradients = (y_pred.data - y_true.data) / avg
    if y_true.requires_grad:
        if y_true.grad is None:
            y_true.grad = gradients
        else:
            y_true.grad += gradients

    if before_softmax_y_pred.requires_grad:
        if before_softmax_y_pred.grad is None:
            before_softmax_y_pred.grad = gradients
        else:
            before_softmax_y_pred.grad += gradients


def CrossEntropyBackward(outputs):
    y_pred, y_true = outputs.in_bounds
    # # before softmax
    before_softmax_y_pred = y_pred.in_bounds[0]
    to_sum_dim = reduce(lambda x, y: x * y, y_pred.data.shape[:-1])
    # to_sum_dim = np.prod(y_pred.data.shape[:-1])
    last_dim = y_pred.data.shape[-1]
    n = y_pred.data.shape[0]
    probs = y_pred.data.reshape(-1, last_dim)
    y_flat = y_true.data.reshape(to_sum_dim)
    probs[np.arange(to_sum_dim), y_flat] -= 1
    gradients = probs.reshape(y_pred.data.shape) / n
    # if y_true.requires_grad:
    #     y_true.grad += gradients
    if before_softmax_y_pred.requires_grad:
        if before_softmax_y_pred.grad is None:
            before_softmax_y_pred.grad = gradients
        else:
            before_softmax_y_pred.grad += gradients




np = __import__('cupy')

inputs = None
outputs = None
graph = None


def topological_sort(inputs, outputs):
    """
    Sort generic nodes in topological order using Kahn's Algorithm.
    Returns a list of sorted nodes.
    """
    name_dict = dict()
    G = {}
    graph = []
    outputs = list([outputs])

    layers = list([inputs])
    while len(layers) > 0:
        n = layers.pop(0)
        if n not in G:
            G[n] = {'in': set(), 'out': set()}
        for m in n.out_bounds:
            if m not in G:
                G[m] = {'in': set(), 'out': set()}
            G[n]['out'].add(m)
            G[m]['in'].add(n)
            layers.append(m)

    S = set(list([inputs]))
    while len(S) > 0:
        n = S.pop()
        graph.append(n)
        if n.name is None:
            if n.__class__.__name__ in name_dict:
                name_dict[n.__class__.__name__] += 1
            else:
                name_dict[n.__class__.__name__] = 0
            n.name = n.__class__.__name__.lower() + str(name_dict[n.__class__.__name__])
        if n in outputs:
            continue
        for m in n.out_bounds:

            G[n]['out'].remove(m)
            G[m]['in'].remove(n)
            # if no other incoming edges add to S
            if len(G[m]['in']) == 0:
                S.add(m)

    return graph


def build_graph():
    global graph
    graph = topological_sort(inputs, outputs)
    return graph


def reset_graph():
    global graph
    global inputs
    global outputs
    del inputs, outputs, graph
    inputs = None
    outputs = None
    graph = None



from ..nn.global_graph import np
import copy


class Optimizer:
    def __init__(self, trainable_variables=None, lr=0.1, weight_decay=0.):
        self.trainable_variables = trainable_variables
        self.lr = lr
        self.weight_decay = weight_decay
        self.iterations = 0

    def zero_grad(self):
        for v in self.trainable_variables:
            v.grad = np.zeros_like(v.data)

    def step(self):
        self.iterations += 1


class SGD(Optimizer):
    def step(self):
        for v in self.trainable_variables:
            v.data = v.data - self.lr * v.grad
        super(SGD, self).step()


class Momentum(Optimizer):
    def __init__(self, trainable_variables=None, lr=0.01, weight_decay=0., rho=0.9):
        self.rho = rho
        self.velocity = None
        super(Momentum, self).__init__(trainable_variables=trainable_variables, lr=lr, weight_decay=weight_decay)

    def step(self):
        if self.velocity is None:
            # initialize
            self.velocity = [np.zeros_like(p.data) for p in self.trainable_variables]

        for i, (v, var) in enumerate(zip(self.velocity, self.trainable_variables)):
            v = self.rho * v + (1 - self.rho) * var.grad
            var.data -= self.lr * v
            self.velocity[i] = v

        super(Momentum, self).step()


class RMSprop(Optimizer):
    def __init__(self, trainable_variables=None, lr=0.001, weight_decay=0., rho=0.9, epsilon=1e-7):
        self.rho = rho
        self.epsilon = epsilon
        self.ms = None
        super(RMSprop, self).__init__(trainable_variables=trainable_variables, lr=lr, weight_decay=weight_decay)

    def step(self):
        if self.ms is None:
            # initialize
            self.ms = [np.zeros_like(p.data) for p in self.trainable_variables]

        for i, (s, var) in enumerate(zip(self.ms, self.trainable_variables)):
            new_s = self.rho * s + (1 - self.rho) * np.square(var.grad)
            var.data -= self.lr * var.grad / np.sqrt(new_s + self.epsilon)
            self.ms[i] = new_s

        super(RMSprop, self).step()


class AdaGrad(Optimizer):
    def __init__(self, trainable_variables=None, lr=0.01, weight_decay=0., epsilon=1e-7):
        self.epsilon = epsilon
        self.ms = None
        super(AdaGrad, self).__init__(trainable_variables=trainable_variables, lr=lr, weight_decay=weight_decay)

    def step(self):
        if self.ms is None:
            self.ms = [np.zeros_like(g.data) for g in self.trainable_variables]
        for i, (s, var)in enumerate(zip(self.ms, self.trainable_variables)):
            s += np.power(var.grad, 2)
            var.data -= self.lr * var.grad / np.sqrt(s + self.epsilon)
            self.ms[i] = s
        super(AdaGrad, self).step()


class AdaDelta(Optimizer):
    def __init__(self, trainable_variables=None, lr=1.0, weight_decay=0.0, rho=0.95, epsilon=1e-7):
        self.rho = rho
        self.epsilon = epsilon
        self.ms = None
        self.delta_x = None
        super(AdaDelta, self).__init__(trainable_variables=trainable_variables, lr=lr, weight_decay=weight_decay)

    def step(self):
        if self.ms is None:
            self.ms = [np.zeros_like(g.grad) for g in self.trainable_variables]
        if self.delta_x is None:
            self.delta_x = [np.zeros_like(g.grad) for g in self.trainable_variables]

        for i, (s, var, x) in enumerate(zip(self.ms, self.trainable_variables, self.delta_x)):
            s = self.rho * s + (1 - self.rho) * np.power(var.grad, 2)
            g_ = np.sqrt((x + self.epsilon) / (s + self.epsilon)) * var.grad
            var.data -= g_
            x = self.rho * x + (1 - self.rho) * np.power(g_, 2)
            self.ms[i] = s
            self.delta_x[i] = x
        super(AdaDelta,self).step()


class Adam(Optimizer):
    def __init__(self, trainable_variables=None, lr=0.001, weight_decay=0.0, beta1=0.9, beta2=0.999, epsilon=1e-7):
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        self.ms = None
        self.vs = None
        super(Adam, self).__init__(trainable_variables=trainable_variables, lr=lr, weight_decay=weight_decay)

    def step(self):
        self.iterations += 1
        if self.ms is None:
            # initialize
            self.ms = [np.zeros_like(p.data) for p in self.trainable_variables]
        if self.vs is None:
            # initialize
            self.vs = [np.zeros_like(p.data) for p in self.trainable_variables]

        for i, (v, m, var) in enumerate(zip(self.vs, self.ms, self.trainable_variables)):
            v = self.beta1 * v + (1 - self.beta1) * var.grad
            m = self.beta2 * m + (1 - self.beta2) * np.square(var.grad)
            v_correct = v / (1 - pow(self.beta1, self.iterations))
            m_correct = m / (1 - pow(self.beta2, self.iterations))
            var.data -= self.lr * (v_correct / (np.sqrt(m_correct) + self.epsilon))
            self.ms[i] = m
            self.vs[i] = v

        super(Adam, self).step()


def get_optimizer(optimizer, **kwargs):
    trainable_variables = kwargs.pop('trainable_variables', None)
    lr = kwargs.pop('lr', 0.1)
    weight_decay = kwargs.pop('weight_decay', 0.)
    if optimizer.__class__.__name__ == 'str':
        optimizer = optimizer.lower()
        if optimizer == 'sgd':
            return SGD(trainable_variables=trainable_variables, lr=lr, weight_decay=weight_decay)
        elif optimizer == 'adam':
            return Adam(trainable_variables=trainable_variables, lr=lr, weight_decay=weight_decay)
        elif optimizer == 'rmsprop':
            return RMSprop(trainable_variables=trainable_variables, lr=lr, weight_decay=weight_decay)
        elif optimizer == 'momentum':
            return Momentum(trainable_variables=trainable_variables, lr=lr, weight_decay=weight_decay)
        elif optimizer == 'adagrad':
            return AdaGrad(trainable_variables=trainable_variables, lr=lr, weight_decay=weight_decay)
        elif optimizer == 'adadelta':
            return AdaDelta(trainable_variables=trainable_variables, lr=lr, weight_decay=weight_decay)
    elif isinstance(optimizer, Optimizer):
        return copy.deepcopy(optimizer)
    else:
        raise ValueError('unknown optimizer type!')



from ..nn.core import Layer, Variable
from ..nn.global_graph import np
from functools import reduce
from ..utils.initializers import get_initializer
from ..nn.functional import dropout2d, batchnorm_2d
from ..nn.grad_fn import Dropout2DBackward, Batchnorm2DBackward


class Dropout(Layer):
    def __init__(self, keep_prob: float):
        # prob :probability of keeping a unit active.
        self.keep_prob = keep_prob
        self.mask = None
        super(Dropout, self).__init__()

    def __call__(self, inbound):
        if isinstance(inbound, Variable):
            output = dropout2d(self.input_data, self.keep_prob)
            # output是一个Variable
            return output
        super(Dropout, self).__call__(inbound)
        return self

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x
        self.data = dropout2d(self.input_data, self.keep_prob)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        Dropout2DBackward(self.data)


class BatchNormalization(Layer):
    def __init__(self, epsilon=1e-6, momentum=0.99, axis=1, gamma_initializer='ones', beta_initializer='zeros',
                 moving_mean_initializer='zeros', moving_variance_initializer='ones', **kwargs):
        # axis=1 when input Fully Connected Layers(data shape:(M,N),where M donotes Batch-size,and N represents feature nums)  ---also axis=-1 is the same
        # axis=1 when input Convolution Layers(data shape:(M,C,H,W),represents Batch-size,Channels,Height,Width,respectively)
        self.epsilon = epsilon
        self.axis = axis
        self.momentum = momentum
        self.gamma_initializer = get_initializer(gamma_initializer)
        self.beta_initializer = get_initializer(beta_initializer)
        self.moving_mean_initializer = get_initializer(moving_mean_initializer)
        self.moving_variance_initializer = get_initializer(moving_variance_initializer)
        self.moving_mean = None
        self.moving_variance = None
        self.cache = None
        super(BatchNormalization, self).__init__(**kwargs)

    def initial_params(self, input_shape=None):
        if input_shape is not None:
            self.input_shape = input_shape
        assert len(self.input_shape) >= 1
        n_in = self.input_shape[self.axis - 1]
        gamma = Variable(self.gamma_initializer(n_in))
        beta = Variable(self.beta_initializer(n_in))
        self.variables.append(gamma)
        self.variables.append(beta)
        self.moving_mean = self.moving_mean_initializer(n_in)
        self.moving_variance = self.moving_variance_initializer(n_in)

    def __call__(self, inbound):
        if isinstance(inbound, Variable):
            if len(self.variables) == 0:
                self.initial_params(inbound.shape[1:])
            output = self.forward(inbound)
            return output
        super().__call__(inbound)
        return self

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x

        self.data = batchnorm_2d(self.input_data, self.variables[0], self.variables[1], self.axis, self.epsilon,
                                 is_training, self.momentum, self.moving_mean, self.moving_variance)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        Batchnorm2DBackward(self.data)


class LayerNormalization(Layer):
    def __init__(self, epsilon=1e-10, gamma_initializer='ones', beta_initializer='zeros'):
        self.epsilon = epsilon
        self.gamma_initializer = get_initializer(gamma_initializer)
        self.beta_initializer = get_initializer(beta_initializer)
        self.shape_field = None
        self.cache = None
        super(LayerNormalization, self).__init__()

    def initial_params(self, input_shape=None):
        gamma = Variable(self.gamma_initializer(self.input_shape))
        beta = Variable(self.beta_initializer(self.input_shape))
        self.variables.append(gamma)
        self.variables.append(beta)

    def __call__(self, inbound):
        Layer.__call__(self, inbound)
        return self

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x
        inputs = self.input_data.data
        self.shape_field = tuple([i for i in range(1, inputs.ndim)])
        gamma, beta = self.variables
        #calc mean
        mean = np.mean(inputs, axis=self.shape_field, keepdims=True)
        #calc var
        var = np.var(inputs, axis=self.shape_field, keepdims=True)

        #x minus u
        xmu = inputs - mean
        sqrtvar = np.sqrt(var + self.epsilon)
        normalized_x = xmu / sqrtvar
        outputs = gamma.data * normalized_x + beta.data
        self.cache = (xmu, sqrtvar, normalized_x)
        self.data = Variable(data=outputs)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        grad = self.data.grad
        xmu, sqrtvar, normalized_x = self.cache
        std_inv = 1. / sqrtvar

        N = reduce(lambda x, y: x * y, normalized_x.shape[1:])
        gamma, beta = self.variables
        if gamma.requires_grad:
            gamma.grad += np.sum(grad * normalized_x, axis=0)
        if beta.requires_grad:
            beta.grad += np.sum(grad, axis=0)

        dnormalized_x = grad * gamma.data
        dvar = (-0.5) * np.sum(dnormalized_x * xmu, axis=self.shape_field, keepdims=True) * (std_inv ** 3) # (m,1)=(m,c,h,w)*(m,c,h,w)*(m,1)

        dmean = (-1.0) * np.sum(dnormalized_x * std_inv, axis=self.shape_field, keepdims=True) - 2.0 * dvar * np.mean(xmu, axis=self.shape_field, keepdims=True)

        outputs = dnormalized_x * std_inv + (2. / N) * dvar * xmu + (1. / N) * dmean

        if self.input_data.requires_grad:
            self.input_data.grad += outputs


class GroupNormalization(Layer):
    def __init__(self, epsilon=1e-5, groups=16, gamma_initializer='ones', beta_initializer='zeros'):
        self.epsilon = epsilon
        self.G = groups
        self.gamma_initializer = get_initializer(gamma_initializer)
        self.beta_initializer = get_initializer(beta_initializer)
        self.shape_field = None
        self.cache = None
        super(GroupNormalization, self).__init__()

    def initial_params(self, input_shape=None):
        c = self.input_shape[0]
        assert c % self.G == 0
        gamma = Variable(self.gamma_initializer((1, c, 1, 1)))
        beta = Variable(self.beta_initializer((1, c, 1, 1)))
        self.variables.append(gamma)
        self.variables.append(beta)

    def __call__(self, inbound):
        Layer.__call__(self, inbound)
        return self

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x

        inputs = self.input_data.data
        gamma, beta = self.variables
        n, c, h, w = inputs.shape
        self.shape_field = tuple([i for i in range(2, inputs.ndim)])

        x_group = np.reshape(inputs, (n, self.G, c // self.G, h, w))
        mean = np.mean(x_group, axis=self.shape_field, keepdims=True)
        var = np.var(x_group, axis=self.shape_field, keepdims=True)
        xgmu = x_group - mean
        sqrtvar = np.sqrt(var + self.epsilon)
        x_group_norm = xgmu / sqrtvar

        x_norm = np.reshape(x_group_norm, (n, c, h, w))

        outputs = gamma.data * x_norm + beta.data

        self.cache = (xgmu, sqrtvar, x_norm)
        self.data = Variable(data=outputs)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        grad = self.data.grad
        xgmu, sqrtvar, x_norm = self.cache
        std_inv = 1. / sqrtvar
        gamma, beta = self.variables
        n, c, h, w = grad.shape
        if gamma.requires_grad:
            gamma.grad += np.sum(grad * x_norm, axis=(0, 2, 3), keepdims=True)
        if beta.requires_grad:
            beta.grad += np.sum(grad, axis=(0, 2, 3), keepdims=True)

        # dx_group_norm
        dx_norm = grad * gamma.data  # (N,C,H,W)
        dx_group_norm = np.reshape(dx_norm, (n, self.G, c // self.G, h, w))
        # dvar
        dvar = -0.5 * (std_inv ** 3) * np.sum(dx_group_norm * xgmu, axis=(2, 3, 4), keepdims=True)
        # dmean
        N_GROUP = c // self.G * h * w
        dmean1 = np.sum(dx_group_norm * -std_inv, axis=(2, 3, 4), keepdims=True)
        dmean2_var = dvar * -2.0 / N_GROUP * np.sum(xgmu, axis=(2, 3, 4), keepdims=True)
        dmean = dmean1 + dmean2_var
        # dx_group
        dx_group1 = dx_group_norm * std_inv
        dx_group_var = dvar * 2.0 / N_GROUP * xgmu
        dx_group_mean = dmean * 1.0 / N_GROUP
        dx_group = dx_group1 + dx_group_var + dx_group_mean
        # dx
        outputs = np.reshape(dx_group, (n, c, h, w))

        if self.input_data.requires_grad:
            self.input_data.grad += outputs



from ..nn.global_graph import np
from typing import List, Tuple


class ProgressBar:
    def __init__(self,
                 max_iter: int = 1,
                 verbose: int = 1,
                 bar_nums: int = 20,
                 untrained_sign: str = '*',
                 trained_sign: str = '='):
        self.max_iter = max_iter
        self.verbose = verbose
        self._nums = bar_nums - 1
        self._untrained = untrained_sign
        self._trained = trained_sign
        self.iter = 0
        self.times = 0

    def update(self, n_iter: int = 1):
        self.iter += n_iter
        self.times += 1

    def get_bar(self) -> str:
        trained_ratio = self.iter / self.max_iter
        reached_bar_nums = round(trained_ratio * self._nums)
        unreached_bar_nums = self._nums - reached_bar_nums
        if self.verbose == 1:
            bar = reached_bar_nums * self._trained + '>' + unreached_bar_nums * self._untrained
        else:
            percent = str(round(trained_ratio * 100))
            bar = '{black} {percent:>{white}}%'.format(black="\033[40m%s\033[0m" % ' ' * reached_bar_nums,
                                                       percent=percent, white=unreached_bar_nums)
        return bar

    def console(self,
                verbose: int = 0,
                trained_time: float = 0.,
                batch_loss: float = 0.,
                batch_acc: float = 0.,
                validation_loss: float = None,
                validation_acc: float = None):

        if verbose == 0:
            return

        bar = self.get_bar()
        if verbose == 1:
            formated_trained_time = format_time(trained_time)
            formated_per_batch_time = format_time(trained_time / self.times)
            if validation_loss == 0. and validation_acc == 0.:
                print('\r {:d}/{:d} [{}] - {} - {}/batch -batch_loss: {:.4f} -batch_acc: {:.4f}'.format(self.iter,
                                                                                                        self.max_iter,
                                                                                                        bar,
                                                                                                        formated_trained_time,
                                                                                                        formated_per_batch_time,
                                                                                                        batch_loss,
                                                                                                        batch_acc),
                      flush=True, end='')
            else:
                print('\r {:d}/{:d} [{}] - {} - {}/batch'
                      ' -batch_loss: {:.4f} -batch_acc: {:.4f} -validation_loss: {:.4f} -validation_acc: {:.4f}'.format(
                    self.iter, self.max_iter, bar, formated_trained_time, formated_per_batch_time, batch_loss,
                    batch_acc, validation_loss, validation_acc), flush=True, end='')
        elif verbose == 2:
            sample_time = trained_time / self.iter
            eta = (self.max_iter - self.iter) * sample_time
            formated_eta = format_time(eta)
            if validation_loss == 0. and validation_acc == 0.:
                print('{} -ETA {} -batch_loss: {:.4f} -batch_acc: {:.4f}'.format(bar, formated_eta, batch_loss,
                                                                                 batch_acc))
            else:
                print(
                    '{} -ETA {} -batch_loss: {:.4f} -batch_acc: {:.4f} -validation_loss: {:.4f} -validation_acc: {:.4f}'.format(
                        bar, formated_eta, batch_loss, batch_acc, validation_loss, validation_acc))
        else:
            raise ValueError('Verbose only supports for 0, 1 and 2 ~')


class AverageMeter(object):
    def __init__(self, name=None, verbose=0):
        self.name = name
        self.val = None
        self.avg = None
        self.sums = None
        self.steps = 0
        self.verbose = verbose
        self.reset()

    def reset(self):
        if self.verbose == 0:
            self.val = 0.
            self.avg = 0.
            self.sums = 0.
        else:
            self.val = []
            self.avg = []
            self.sums = []

    def update(self, val, step=1):
        if val is None:
            self.val = None
            return
        self.steps += step
        if self.verbose == 0:
            self.val = val
            self.sums += val * step
            self.avg = self.sums / self.steps
        else:
            self.val.append(val)
            self.sums.append(self.sums[-1] + val * step)
            self.avg.append(self.sums[-1] / self.steps)


class DataSet:
    def __init__(self, *datas):
        self.datas = list(datas)

    def __len__(self):
        return len(self.datas[0])

    def __getitem__(self, item):
        ret_list = []
        for data in self.datas:
            ret_list.append(data[item])
        return ret_list


class DataLoader:
    def __init__(self, dataset: DataSet, batch_size: int, shuffle: bool = False, seed: int = None):
        self.dataset = dataset
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.seed = 0 if seed is None else seed
        self.mini_batches = self.make_batches(self.dataset.datas, self.batch_size, self.seed, self.shuffle)
        self.sp = -1

    def __iter__(self):
        return self

    def __next__(self):
        self.sp += 1
        if self.sp >= len(self.mini_batches):
            self.seed += 1
            self.mini_batches = self.make_batches(self.dataset.datas, self.batch_size, self.seed, self.shuffle)
            self.sp = -1
            raise StopIteration
        return self.mini_batches[self.sp]

    def make_batches(self, datas: List[np.ndarray], batch_size: int, seed: int, shuffle: bool = False):
        np.random.seed(seed)
        m = datas[0].shape[0]
        if shuffle:
            permutation = np.random.permutation(m)
            datas = list(map(lambda x: x[permutation], datas))

        mini_batches = []
        complete_batch_nums = m // batch_size  # 完整的mini_batch个数
        for i in range(complete_batch_nums):
            mini_batch = list(map(lambda x: x[batch_size * i:batch_size * (i + 1)], datas))
            mini_batches.append(mini_batch)

        if m % batch_size != 0:
            mini_batches.append(list(map(lambda x: x[batch_size * complete_batch_nums:], datas)))
        return mini_batches


def format_time(second_time: float) -> str:
    if second_time < 1:
        ms = second_time * 1000
        if ms < 1:
            us = second_time * 1000
            return '%dus' % us
        else:
            return '%dms' % ms
    second_time = round(second_time)
    if second_time > 3600:
        # hours
        h = second_time // 3600
        second_time = second_time % 3600
        # minutes
        m = second_time // 60
        second_time = second_time % 60
        return '%dh%dm%ds' % (h, m, second_time)
    elif second_time > 60:
        m = second_time // 60
        second_time = second_time % 60
        return '%dm%ds' % (m, second_time)
    else:
        return '%ds' % second_time


def im2col(inputs: np.ndarray, out_h: int, out_w: int, kernel_h: int, kernel_w: int, stride: Tuple):
    batch_nums, n_C_prev, n_H_prev, n_W_prev = inputs.shape
    col = np.zeros((batch_nums, n_C_prev, kernel_h, kernel_w, out_h, out_w))

    for y in range(kernel_h):
        y_max = y + stride[0] * out_h
        for x in range(kernel_w):
            x_max = x + stride[1] * out_w
            col[:, :, y, x, :, :] = inputs[:, :, y:y_max:stride[0], x:x_max:stride[1]]

    col = col.transpose((0, 4, 5, 1, 2, 3)).reshape(batch_nums * out_h * out_w, -1)
    return col


def col2im(inputs_shape: tuple, pad_size: int, kernel_h: int, kernel_w: int, stride: Tuple, dcol: np.ndarray):
    batch_nums, n_C_prev, n_H_prev, n_W_prev = inputs_shape  # 填充前的shape
    n_H = (n_H_prev + 2 * pad_size - kernel_h) // stride[0] + 1
    n_W = (n_W_prev + 2 * pad_size - kernel_w) // stride[1] + 1

    dcol = dcol.reshape((batch_nums, n_H, n_W, n_C_prev, kernel_h, kernel_w)).transpose(0, 3, 4, 5, 1, 2)

    output = np.zeros(
        (batch_nums, n_C_prev, n_H_prev + 2 * pad_size + stride[0] - 1, n_W_prev + 2 * pad_size + stride[1] - 1))

    for y in range(kernel_h):
        y_max = y + stride[0] * n_H
        for x in range(kernel_w):
            x_max = x + stride[1] * n_W
            output[:, :, y:y_max:stride[0], x:x_max:stride[1]] += dcol[:, :, y, x, :, :]

    return output[:, :, pad_size:n_H_prev + pad_size, pad_size:n_W_prev + pad_size]


def initialize_ops_grad(*ops):
    # 初始化梯度
    for op in ops:
        if op is not None and op.requires_grad:
            op.grad = np.zeros_like(op.data)



from ..nn.core import Variable
from ..nn.grad_fn import *
from ..utils.toolkit import im2col, initialize_ops_grad
from ..utils.initializers import Zeros, Ones
from typing import Tuple


def relu(inputs: Variable, inplace: bool = False) -> Variable:
    if inplace:
        inputs.data[inputs.data < 0] = 0
        return inputs
    outputs = Variable(data=np.maximum(0, inputs.data), in_bounds=[inputs])
    outputs.grad_fn = ReluBackward
    inputs.out_bounds.append(outputs)
    initialize_ops_grad(inputs)
    return outputs


def sigmoid(inputs: Variable):
    outputs = 1. / (1 + np.exp(-inputs.data))
    outputs = Variable(data=outputs, in_bounds=[inputs])
    outputs.grad_fn = SigmoidBackward
    inputs.out_bounds.append(outputs)
    initialize_ops_grad(inputs)
    return outputs


def tanh(inputs: Variable):
    outputs = np.tanh(inputs.data)
    outputs = Variable(data=outputs, in_bounds=[inputs])
    outputs.grad_fn = TanhBackward
    inputs.out_bounds.append(outputs)
    initialize_ops_grad(inputs)
    return outputs


def softmax(inputs: Variable):
    # more stable softmax
    shiftx = inputs.data - np.max(inputs.data)
    outputs = np.divide(np.exp(shiftx), np.sum(np.exp(shiftx), axis=-1, keepdims=True))
    outputs = Variable(data=outputs, in_bounds=[inputs])
    inputs.out_bounds.append(outputs)
    initialize_ops_grad(inputs)
    return outputs


def dense(inputs: Variable, weight: Variable, bias: Variable = None):
    outputs = inputs.matmul(weight)
    if bias is not None:
        outputs += bias
    else:
        # 这里加入None是为了数组大小能对应
        outputs.in_bounds.append(None)
    outputs.grad_fn = DenseBackward
    initialize_ops_grad(inputs, weight, bias)
    return outputs


def flatten(inputs: Variable, start: int = 1,inplace: bool = False):
    output_shape = tuple(inputs.shape[:start]) + (-1, )
    if inplace:
        inputs.data = inputs.data.reshape(output_shape)
        return inputs
    outputs = Variable(data=inputs.data.reshape(output_shape), in_bounds=[inputs])
    outputs.grad_fn = FlattenBackward
    inputs.out_bounds.append(outputs)
    initialize_ops_grad(inputs)
    return outputs


def conv2d(inputs: Variable, weight: Variable, bias: Variable = None, stride: Tuple = 1, padding: int = 0):
    # before pad size
    batch_nums, n_c_prev, n_h_prev, n_w_prev = inputs.data.shape
    # pad
    pad_data = np.pad(inputs.data, ((0, 0), (0, 0), (padding, padding), (padding, padding)), 'constant')

    out_channels, in_channels, kernel_h, kernel_w = weight.data.shape
    # output size
    n_h = (n_h_prev - kernel_h + 2 * padding) // stride[0] + 1
    n_w = (n_w_prev - kernel_w + 2 * padding) // stride[1] + 1
    col = im2col(pad_data, n_h, n_w, kernel_h, kernel_w, stride)

    col_w = weight.data.reshape(out_channels, -1).T

    outputs = col.dot(col_w)
    if bias is not None:
        outputs += bias.data
    outputs = outputs.reshape(batch_nums, n_h, n_w, -1).transpose(0, 3, 1, 2)
    outputs = Variable(data=outputs, in_bounds=[inputs, weight, bias])
    # store these for bp
    outputs.cache['col'] = col
    outputs.cache['stride'] = stride
    outputs.cache['pad_size'] = padding
    outputs.grad_fn = Conv2DBackward
    inputs.out_bounds.append(outputs)
    initialize_ops_grad(inputs, weight, bias)
    return outputs


def max_pool2d(inputs: Variable, kernel_size: int = 2, stride: Tuple = None, padding: int = 0):
    if stride is None:
        stride = kernel_size
    if padding != 0:
        data = np.pad(inputs.data, ((0, 0), (0, 0), (padding, padding), (padding, padding)), 'constant')
    else:
        data = inputs.data

    n, c, h, w = data.shape
    if kernel_size == stride:
        mode = 'reshape'

        x_reshaped = data.reshape((n, c, h // kernel_size, kernel_size, w // kernel_size, kernel_size))
        outputs = Variable(data=x_reshaped.max(axis=3).max(axis=4), in_bounds=[inputs])
        outputs.cache['x_reshaped'] = x_reshaped

    else:
        mode = 'im2col'

        out_h, out_w = (h - kernel_size) // stride[0] + 1, (w - kernel_size) // stride[1] + 1

        col = im2col(data, out_h, out_w, kernel_size, kernel_size, stride)
        col = col.reshape(-1, kernel_size * kernel_size)
        pool_argmax = np.argmax(col, axis=1)
        outputs = np.max(col, axis=1).reshape((n, out_h, out_w, c)).transpose(0, 3, 1, 2)
        outputs = Variable(data=outputs, in_bounds=[inputs])
        outputs.cache['pool_argmax'] = pool_argmax
        outputs.cache['kernel_size'] = kernel_size

    outputs.cache['mode'] = mode
    outputs.cache['pad_size'] = padding
    outputs.cache['stride'] = stride
    outputs.grad_fn = Maxpool2DBackward
    inputs.out_bounds.append(outputs)
    initialize_ops_grad(inputs)
    return outputs


def avg_pool2d(inputs: Variable, kernel_size: int, stride: Tuple = None, padding: int = 0):
    if padding != 0:
        data = np.pad(inputs.data, ((0, 0), (0, 0), (padding, padding), (padding, padding)), 'constant')
    else:
        data = inputs.data

    n, c, h, w = data.shape
    out_h, out_w = (h - kernel_size) // stride[0] + 1, (w - kernel_size) // stride[1] + 1
    col = im2col(data, out_h, out_w, kernel_size, kernel_size, stride)
    col = col.reshape(-1, kernel_size * kernel_size)
    pool_argmean = np.array([range(col.shape[1])])
    outputs = np.mean(col, axis=1).reshape((n, out_h, out_w, c)).transpose(0, 3, 1, 2)
    outputs = Variable(data=outputs, in_bounds=[inputs])
    outputs.cache['pool_argmean'] = pool_argmean
    outputs.cache['kernel_size'] = kernel_size
    outputs.cache['pad_size'] = padding
    outputs.cache['stride'] = stride
    outputs.grad_fn = Avgpool2DBackward
    inputs.out_bounds.append(outputs)
    initialize_ops_grad(inputs)
    return outputs


def pad_2d(inputs: Variable, padding: tuple, inplace: bool = False):
    if inplace:
        inputs.data = np.pad(inputs.data, ((0, 0), (0, 0), (padding[0], padding[0]), (padding[1], padding[1])),
                             'constant')
        return inputs
    outputs = Variable(data=np.pad(inputs.data, ((0, 0), (0, 0), (padding[0], padding[0]), (padding[1], padding[1])),
                                   'constant'), in_bounds=[inputs])
    outputs.cache['pad_size'] = padding
    outputs.grad_fn = Pad2DBackward
    inputs.out_bounds.append(outputs)
    initialize_ops_grad(inputs)
    return outputs


def dropout2d(inputs: Variable, keep_prob: float = 0.5):
    random_tensor = np.random.binomial(n=1, p=keep_prob, size=inputs.data.shape)
    outputs = inputs.data * random_tensor / keep_prob
    outputs = Variable(data=outputs, in_bounds=[inputs])
    outputs.cache['mask'] = random_tensor
    outputs.cache['keep_prob'] = keep_prob
    outputs.grad_fn = Dropout2DBackward
    inputs.out_bounds.append(outputs)
    initialize_ops_grad(inputs)
    return outputs


def batchnorm_2d(inputs: Variable, gamma: Variable, beta: Variable, axis: int, epsilon: float, training: bool,
                 momentum: float = 0.99, moving_mean: np.ndarray = None, moving_variance: np.ndarray = None):
    if moving_mean is not None:
        inputs.cache['moving_mean'] = moving_mean
    if moving_variance is not None:
        inputs.cache['moving_variance'] = moving_variance
    if 'moving_mean' not in inputs.cache.keys():
        inputs.cache['moving_mean'] = Zeros()(inputs.shape[axis])
    if 'moving_variance' not in inputs.cache.keys():
        inputs.cache['moving_variance'] = Ones()(inputs.shape[axis])

    inputs_data = inputs.data
    ndim = inputs_data.ndim

    if not (axis == -1 or axis == ndim - 1):
        inputs_data = np.swapaxes(inputs_data, axis, -1)

    before_reshape_shape = inputs_data.shape
    inputs_data = inputs_data.reshape(-1, inputs.data.shape[axis])

    if training:
        # calc mean
        mean = np.mean(inputs_data, axis=0)
        # calc var
        var = np.var(inputs_data, axis=0)
        # x minus u
        xmu = inputs_data - mean
        sqrtvar = np.sqrt(var + epsilon)
        normalized_x = xmu / sqrtvar
        outputs = gamma.data * normalized_x + beta.data

        inputs.cache['xmu'] = xmu
        inputs.cache['axis'] = axis
        inputs.cache['sqrtvar'] = sqrtvar
        inputs.cache['normalized_x'] = normalized_x
        inputs.cache['moving_mean'] = momentum * inputs.cache['moving_mean'] + (1 - momentum) * mean
        inputs.cache['moving_variance'] = momentum * inputs.cache['moving_variance'] + (1 - momentum) * var

    else:
        scale = gamma.data / (np.sqrt(inputs.cache['moving_variance'] + epsilon))
        outputs = inputs * scale + (beta.data - inputs.cache['moving_mean'] * scale)

    outputs = outputs.reshape(before_reshape_shape)

    if not (axis == -1 or axis == ndim - 1):
        # for instance,outputs:(N,W,H,C), self.axis=1, after swapaxes,outputs:(N,C,H,W)
        outputs = np.swapaxes(outputs, axis, -1)

    outputs = Variable(data=outputs, in_bounds=[inputs, gamma, beta])
    outputs.grad_fn = Batchnorm2DBackward
    initialize_ops_grad(inputs)
    inputs.out_bounds.append(outputs)
    return outputs





def concatenate(*variables: Variable, axis: int, output: Variable = None, name: str = None):
    data = variables[0].data
    for i in range(1, len(variables)):
        data = np.concatenate((data, variables[i].data), axis=axis)
    if output is None:
        output = Variable(data=data, name=name, in_bounds=[variables])
    else:
        output.data = data
        output.shape = data.shape
        output.name = name
        output.in_bounds.append(*variables)
    return output



from ..nn.grad_fn import *
from typing import List, Tuple, Union
import gc
from ..nn import global_graph as GlobalGraph
from ..utils.toolkit import initialize_ops_grad


# 第一种数据类型Node，也就是我们基本的tensor type，每一个tensor都是Node的实例化类型
class Node:
    def __init__(self, in_bounds: List = None, out_bounds: List = None, data: GlobalGraph.np.ndarray = None,
                 shape: Union[List, Tuple] = None, name: str = None, requires_grad: bool = False):
        """
        :param in_bounds: 与当前tensor相连的Node或者是Layer，需要重载下：一.可能是Layer数组，二.可能是Node数组（主要以Variable为主）
        :param out_bounds: 输出的Node或Layer，也是需要重载，Layer数组或者Node数组
        :param data: 一个numpy矩阵
        :param name: tensor的名字，字符串
        :param requires_grad: 需不需要梯度， 布尔值
        """

        # 下面第一行这句话的意思是如果传递了in_bounds参数，那就把传递的in_bounds先包装为数组，在赋给类成员in_bounds，否则的话初始化类成员in_bounds为空数组，如果c++写的话大概是这样：
        # this->in_bounds = in_bounds == NULL ?  vector<Template> : vector<Template>(in_bounds)
        self.in_bounds = [] if in_bounds is None else list(in_bounds)
        self.out_bounds = [] if out_bounds is None else list(out_bounds)
        self.data = data
        self.shape = self.__check_shape(shape)
        self.name = name
        self.requires_grad = requires_grad
        # grad也是一个numpy矩阵，默认为None，之后会初始化它，在cpp中这样写？
        # 先声明ndarray grad,然后this->grad = NULL
        self.grad = None
        # tensor求梯度的函数，这个要绑定一个函数，比如绑定的是AddBackward（self.grad_fn = AddBackward），那么调用self.grad_fn和调用AddBackward一样

        self.grad_fn = None
        # 存储一些必要的数据用于反向传播，类型为一个字典，通过key-value形式索引
        self.cache = {}

    # 该方法相当于重载对类的print, info就是print类后显示的结果
    def __repr__(self) -> str:
        # {}里填写的变量就相当于是把这个变量的值转换为字符串填在这
        if self.grad_fn:
            info = f'{self.__class__.__name__}({self.data}, requires_grad={self.requires_grad}, ' \
                   f'grad_fn={self.grad_fn.__name__})'
        else:
            info = f'{self.__class__.__name__}({self.data}, requires_grad={self.requires_grad}, ' \
                   f'grad_fn={self.grad_fn})'
        return info

    def __str__(self) -> str:
        # 同上，不在解释，在python里要重载类的print，最好是把这两个方法都重载了
        if self.grad_fn:
            info = f'{self.__class__.__name__}({self.data}, requires_grad={self.requires_grad}, ' \
                   f'grad_fn=<{self.grad_fn.__name__}>)'
        else:
            info = f'{self.__class__.__name__}({self.data}, requires_grad={self.requires_grad})'
        return info

    def __check_shape(self, shape: Union[List, Tuple]) -> Union[List, Tuple]:
        # 这个是numpy矩阵问题，当一个numpy array存储的是标量时，默认shape是空，因此要手动设为1
        if len(shape) == 0:
            return tuple([1])
        return shape

    # 重载 + - * / 等运算符， 先默认所有的操作数都是Variable类型吧，之后要重载其他类型再说
    # 加法
    def __add__(self, other):
        # 全局运算图
        if GlobalGraph.inputs is None:
            GlobalGraph.inputs = self
        return add(self, other)

    # 加等于
    def __iadd__(self, other):
        if GlobalGraph.inputs is None:
            GlobalGraph.inputs = self

        self.data += other.data
        self.in_bounds.append(other)
        other.out_bounds.append(self)
        if 'grad_fn' not in self.cache:
            self.cache['grad_fn'] = []
        self.cache['grad_fn'].append(self.grad_fn)
        self.grad_fn = IAddBackward
        return self

    # 取负号
    def __neg__(self):
        return neg(self)

    # 减法
    def __sub__(self, other):
        # a - b = a + (-b)
        return add(self, neg(other))

    # 乘法
    def __mul__(self, other):
        if GlobalGraph.inputs is None:
            GlobalGraph.inputs = self
        return mul(self, other)

    # 幂
    def __pow__(self, power: int, modulo=None):
        outputs = Variable(in_bounds=[self], data=np.power(self.data, power), requires_grad=True)
        # 绑定反向求梯度的函数
        outputs.grad_fn = PowBackward
        outputs.power = power
        initialize_ops_grad(self)
        return outputs

    def backward(self, gradients: GlobalGraph.np.ndarray = None):
        if self.grad_fn is None:
            raise ValueError('can not solve grad on %s' % self)
        # 如果是标量，默认该标量梯度为1
        if self.data.size == 1:
            if self.grad is None:
                self.grad = 1.
            self.grad_fn(self)
            if GlobalGraph.inputs is not None and GlobalGraph.outputs is not None:
                graph = GlobalGraph.build_graph()
                for node in reversed(graph):
                    if node.grad_fn is not None:
                        node.grad_fn(node)
                    if node == GlobalGraph.inputs or node == GlobalGraph.outputs:
                        continue
                    del node.cache, node.data, node.grad, node.grad_fn, node.in_bounds, node.out_bounds, node.shape, \
                        node.name, node.requires_grad
                    del node
                gc.collect()
                # 这里反向传播计算完就把这个图销毁，在C++中就是释放内存,把GlobalGraph的inputs, outputs和graph的内存都释放掉，或者说重置参数
                GlobalGraph.reset_graph()

        # 如果不是标量，但是手动传入该tensor的梯度，也可以求导，但是一定要让传入的梯度shape和tensor的shape对上
        elif gradients is not None:
            self.grad = gradients
            self.grad_fn(self)
        else:
            raise ValueError('grad can be implicitly created only for scalar outputs')

    # 把数据转换为长整型，返回一个Variable
    def long(self):
        output = Variable(data=self.data, dtype=np.int64)
        return output

    # 方法名后面加一个_指在本身数据上直接修改，没有返回值
    def long_(self):
        self.data = self.data.astype(np.int64)

    # 矩阵乘
    def matmul(self, other):
        outputs_data = np.dot(self.data, other.data)
        outputs = Variable(data=outputs_data, in_bounds=[self, other])
        outputs.grad_fn = MatmulBackward
        initialize_ops_grad(self, other)
        self.out_bounds.append(outputs)
        other.out_bounds.append(outputs)
        return outputs

    # 转置
    def t(self):
        outputs = Variable(data=self.data.T, in_bounds=[self])
        outputs.grad_fn = TransposeBackward
        self.out_bounds.append(outputs)
        return outputs

    # 求和
    def sum(self, axis: int = None, keepdims: bool = False):
        # 在哪个axis（维度）上求和
        if axis is None:
            sum_value = np.sum(self.data, keepdims=keepdims)
        else:
            sum_value = np.sum(self.data, axis=axis, keepdims=keepdims)
        outputs = Variable(in_bounds=[self], data=sum_value)
        # 绑定反向求梯度的函数
        outputs.grad_fn = SumBackward
        initialize_ops_grad(self)
        self.out_bounds.append(outputs)
        return outputs

    # 求均值
    def mean(self, axis: int = None, keepdims: bool = False):
        if axis is None:
            mean_value = np.mean(self.data, keepdims=keepdims)
        else:
            mean_value = np.mean(self.data, axis=axis, keepdims=keepdims)
        outputs = Variable(in_bounds=[self], data=mean_value)
        # 绑定反向求梯度的函数
        outputs.grad_fn = MeanBackward
        initialize_ops_grad(self)
        self.out_bounds.append(outputs)
        return outputs

    # 求绝对值
    def abs(self):
        outputs = Variable(in_bounds=[self], data=np.abs(self.data))
        # 绑定反向求梯度的函数
        outputs.grad_fn = AbsBackward
        initialize_ops_grad(self)
        self.out_bounds.append(outputs)
        return outputs

    # reshape，传入的参数应该是一个list或者tuple
    def view(self, *shapes):
        outputs = Variable(in_bounds=[self], data=self.data.reshape(*shapes))
        # 绑定反向求梯度的函数
        outputs.grad_fn = AbsBackward
        initialize_ops_grad(self)
        self.out_bounds.append(outputs)
        return outputs

    # 求log
    def log(self, base: int = 2):
        # 这里可能需要重载一下, base可选为2， 10和'e'
        if base == 2:
            ret_value = np.log2(self.data)
        elif base == 10:
            ret_value = np.log10(self.data)
        else:
            ret_value = np.log(self.data)
        outputs = Variable(in_bounds=[self], data=ret_value)
        # 绑定反向求梯度的函数
        outputs.grad_fn = LogBackward
        # 记录下base供反向传播用
        outputs.cache['base'] = base
        initialize_ops_grad(self)
        self.out_bounds.append(outputs)
        return outputs

    # 求exp
    def exp(self):
        outputs = Variable(in_bounds=[self], data=np.exp(self.data))
        # 绑定反向求梯度的函数
        outputs.grad_fn = ExpBackward
        initialize_ops_grad(self)
        self.out_bounds.append(outputs)
        return outputs


class Variable(Node):
    def __init__(self, data: Union[GlobalGraph.np.ndarray, int, float], in_bounds: List = None, out_bounds: Union[List, Tuple] = None,
                 name: str = None, requires_grad: bool = True, dtype=np.float64):
        # Variable初始化时必须提供data值，data值可能传入的是一个int或者float，我们需要把它包装成numpy矩阵（cpp中就包装成我们选的矩阵库类型）
        data = np.asarray(data, dtype=dtype)  # 数据类型用float64吧，float32也行
        Node.__init__(self,
                      in_bounds=in_bounds,
                      out_bounds=out_bounds,
                      data=data,
                      shape=data.shape,
                      name=name,
                      requires_grad=requires_grad)

    def __getitem__(self, item):
        ret = Variable(data=self.data[item])
        return ret

    def __setitem__(self, key, value):
        self.data[key] = value.data
        self.in_bounds.append(value)


class Constant(Node):
    def __init__(self, data: Union[GlobalGraph.np.ndarray, int, float], in_bounds: List = None, out_bounds: Union[List, Tuple] = None,
                 name: str = None):
        # Constant初始化时必须提供data值，并且一旦初始化就不可修改，Constant因为值不需要修改，也就没必要计算梯度，默认require_grads为False
        # !!!!!!!!!!!!!!!!注意Constant里的这个data要是const类型的，总之就是Constant的data一旦初始化后没办法被修改
        data = np.asarray(data, dtype=np.float64)
        # 显示调用父类的初始化函数
        Node.__init__(self,
                      in_bounds=in_bounds,
                      out_bounds=out_bounds,
                      data=data,
                      name=name,
                      shape=data.shape,
                      requires_grad=False)


class Layer:
    def __init__(self, in_bounds: List = None, out_bounds: List = None, input_shape: Union[List, Tuple] = None,
                 input_data: Variable = None, data: Variable = None, shape: Union[List, Tuple] = None,
                 variables: List[Variable] = None, name: str = None):

        """
        不同于Node，Layer有输入值(input_data，并且是一个Variable)，有运算后的输出值(data，也是一个Variable)，同样有输入数据的shape(input_shape)和输出数据的shape(shape)，还有每一个Layer有自己所做运算需要的参数，如对于全连接层就是w和b，它们存在variables里，variabels的类型是Variable数组
        :param in_bounds:
        :param out_bounds:
        :param input_shape:
        :param shape:
        :param input_value:
        :param data:
        :param variables:
        :param name:
        """
        self.in_bounds = [] if in_bounds is None else list(in_bounds)
        self.out_bounds = [] if out_bounds is None else list(out_bounds)
        self.data = data
        self.input_shape = input_shape
        self.input_data = input_data
        self.name = name
        self.shape = shape
        # 如果传入的variables不是None，就赋值，否则初始化为空的数组
        self.variables = variables if variables else []

    def __call__(self, inbound):
        # 实际inbound可能是Layer也可能是Variable，需要重载，这里在父类里只实现了是Layer的情况，是Variable的情况在子类里在特殊实现。每个子类都会手动调用父类的__call__方法
        # 输入是Layer
        self.shape = self.compute_output_shape(inbound.shape)
        self.in_bounds.append(inbound)
        self.input_shape = inbound.shape
        inbound.out_bounds.append(self)
        return self

    def connect(self, inbound=None):
        # inbound只能是一个Layer
        if inbound is None:
            if self.input_shape is None:
                raise ValueError('must specify input_shape')
        else:
            self.input_shape = inbound.shape
        self.shape = self.compute_output_shape(self.input_shape)
        if inbound is not None:
            self.in_bounds.append(inbound)
            self.input_shape = inbound.shape
            inbound.out_bounds.append(self)

    # 初始化本层参数，由子类单独实现
    def initial_params(self, *args):
        pass

    # 计算本层输出shape，由子类实现
    def compute_output_shape(self, input_shape: Union[List, Tuple] = None) -> Union[List, Tuple]:
        return input_shape

    # 计算本层参数数量，由子类实现
    def params_count(self) -> int:
        total_params = 0
        for v in self.variables:
            if v is not None:
                total_params += v.data.size
        return total_params

    def forward(self, x: Variable = None, is_training: bool = True, *args) -> Variable:
        # x代表本层的输入variable
        raise NotImplemented

    # 做一些forward后的初始化工作，每个子类forward后都会调用该函数
    def connect_init(self, data: Variable, is_training: bool = True):
        for out_bound in self.out_bounds:
            out_bound.input_data = data

        if self.data.requires_grad and is_training:
            initialize_ops_grad(self.data)
            initialize_ops_grad(*self.variables)

    def backward(self, gradients: GlobalGraph.np.ndarray = None):
        for inbound in self.in_bounds:
            if inbound.data.requires_grad:
                inbound.data.grad += gradients


# 加法
def add(*ops: Variable) -> Variable:
    data = 0
    for op in ops:
        data += op.data
    outputs = Variable(in_bounds=[*ops], data=data)
    initialize_ops_grad(ops)
    # 绑定反向求梯度的函数
    outputs.grad_fn = AddBackward
    for op in ops:
        op.out_bounds.append(outputs)
    return outputs


# 对自身取负
def neg(x: Variable) -> Variable:
    outputs = Variable(in_bounds=[x], data=-x.data, requires_grad=True)
    initialize_ops_grad(x)
    # 绑定反向求梯度的函数
    outputs.grad_fn = NegBackward
    return outputs


# 乘法
def mul(x: Variable, y: Variable) -> Variable:
    # 检查下x和y的shape，看看是要做矩阵乘法运算还是点乘运算

    # 列举所有做点乘的情况, 1.两个标量， 2.两个矩阵的shape一样。
    # 有一种特殊情况，两个矩阵的shape虽然不一样，但还是可以通过broadcast机制做点乘，比如(1, 3) * (3, 3) = (3, 3)，但是也可以理解为矩阵乘 = (1, 3)，这种情况下，我们默认为矩阵乘。

    if len(x.shape) == 1 and len(y.shape) == 1 or x.shape == y.shape:
        # 点乘
        outputs = Variable(in_bounds=[x, y], data=x.data * y.data, requires_grad=True)
        outputs.grad_fn = MultiplyBackward
    else:
        # 矩阵乘
        outputs = Variable(in_bounds=[x, y], data=np.dot(x.data, y.data), requires_grad=True)
        outputs.grad_fn = MatmulBackward

    initialize_ops_grad(x, y)
    x.out_bounds.append(outputs)
    y.out_bounds.append(outputs)
    return outputs



from ..nn.global_graph import np


def to_categorical(inputs):
    if inputs.ndim > 2:
        raise ValueError('only accept 1-d or 2-d inputs')
    # convert Y to one-hot encode
    # for example,merge (batch,1) to (batch,)
    if inputs.ndim == 2:
        if inputs.shape[-1] == 1:
            inputs = np.sum(inputs, axis=-1)
        else:
            raise ValueError('can not convert %s to one-hot vector' % (inputs.__class__))

    n_class = np.max(inputs)[0] + 1
    encoded_Y = np.eye(n_class)[inputs].reshape(-1, n_class)
    return encoded_Y


def pad_sequences(sequences, maxlen=None, dtype='int32', padding='pre', truncating='pre', value=0.):
    lengths = []
    for x in sequences:
        lengths.append(len(x))
    num_samples = len(sequences)
    if maxlen is None:
        maxlen = max(lengths)

    sample_shape = tuple()
    for s in sequences:
        if len(s) > 0:
            sample_shape = np.asarray(s).shape[1:]
            break

    x = (np.ones((num_samples, maxlen) + sample_shape) * value).astype(dtype)

    for idx, s in enumerate(sequences):
        if not len(s):
            continue

        if truncating == 'pre':
            trunc = s[-maxlen:]
        elif truncating == 'post':
            trunc = s[:maxlen]
        else:
            raise ValueError('unknown truncating type!')

        trunc = np.asarray(trunc, dtype=dtype)

        if padding == 'pre':
            x[idx, -len(trunc):] = trunc
        elif padding == 'post':
            x[idx, :len(trunc)] = trunc
        else:
            raise ValueError('unknown padding type!')
    return x



from ..nn.core import Layer, Variable
from functools import reduce
from ..utils.initializers import get_initializer
from xshinnosuke.layers.activators import get_activator
from ..nn.functional import dense, flatten
from ..nn.grad_fn import DenseBackward, FlattenBackward
from ..nn import global_graph as GlobalGraph


class Flatten(Layer):
    def __init__(self, start=1, **kwargs):
        # output dimensions
        if start < 1:
            raise ValueError('start must be > 0')
        self.start = start
        super(Flatten, self).__init__(**kwargs)

    def __call__(self, inbound):
        if isinstance(inbound, Variable):
            if GlobalGraph.inputs is None:
                GlobalGraph.inputs = inbound

            output = flatten(inbound, self.start)
            # output是一个Variable
            return output
        super().__call__(inbound)
        return self

    def compute_output_shape(self, input_shape=None):
        assert len(input_shape) >= 3
        flatten_shape = reduce(lambda x, y: x * y, input_shape[self.start - 1:])
        output_shape = input_shape[: self.start - 1] + (flatten_shape,)
        return output_shape

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x
        self.data = flatten(self.input_data, self.start)
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        FlattenBackward(self.data)


class Dense(Layer):
    def __init__(self, out_features, activation=None, use_bias=True, kernel_initializer='normal',
                 bias_initializer='zeros', kernel_regularizer=None, **kwargs):
        self.out_features = out_features
        self.use_bias = use_bias
        self.kernel_initializer = get_initializer(kernel_initializer)
        self.bias_initializer = get_initializer(bias_initializer)
        self.activation = get_activator(activation) if activation is not None else None
        self.kernel_regularizer = kernel_regularizer
        self.timedist_grad = None
        super(Dense, self).__init__(**kwargs)

    def __call__(self, inbound):
        if isinstance(inbound, Variable):
            if GlobalGraph.inputs is None:
                GlobalGraph.inputs = inbound

            if len(self.variables) == 0:
                self.initial_params(inbound.shape[1:])
            output = dense(inbound, self.variables[0], self.variables[1])
            if self.activation is not None:
                output = self.activation.forward(output)
            # output是一个Variable
            return output

        super().__call__(inbound)
        return self

    def initial_params(self, input_shape=None):
        if input_shape is not None:
            self.input_shape = input_shape
        w = Variable(self.kernel_initializer(self.input_shape + (self.out_features, )), name='variable')
        if self.use_bias:
            b = Variable(self.bias_initializer((1, self.out_features)), name='variable')
        else:
            b = None
        self.variables.append(w)
        self.variables.append(b)

    def compute_output_shape(self, input_shape=None):
        return tuple([self.out_features, ])

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x
        w, b = self.variables
        self.data = dense(self.input_data, w, b)
        if self.activation is not None:
            output = self.activation.forward(self.data)
            self.connect_init(output, is_training)
            return output
        else:
            self.connect_init(self.data, is_training)
            return self.data

    def backward(self, gradients=None):
        if self.activation is not None:
            self.activation.backward()
        DenseBackward(self.data)



from ..nn.core import Layer, Variable
from ..utils.initializers import get_initializer
from ..nn.global_graph import np


class Embedding(Layer):
    def __init__(self, input_dim, output_dim, embeddings_initializer='uniform', mask_zero=False, input_length=None, **kwargs):
        super(Embedding, self).__init__(**kwargs)
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.initializer = get_initializer(embeddings_initializer)
        self.mask_zero = mask_zero
        self.input_length = input_length

    def initial_params(self):
        w = Variable(self.initializer((self.input_dim, self.output_dim)), name='embedding_w')
        self.variables.append(w)

    def compute_output_shape(self, input_shape=None):
        return input_shape + (self.output_dim,)

    def __call__(self, inbound):
        assert self.output_dim is not None
        super(Embedding, self).__call__(inbound)

    def forward(self, x: Variable = None, is_training=True, *args):
        if x is not None:
            self.input_data = x
        assert self.input_data.data.ndim == 2
        w, = self.variables
        # to one-hot
        self.data = w.data[self.input_data.data]
        self.connect_init(self.data, is_training)
        return self.data

    def backward(self, gradients=None):
        w, = self.variables
        if w.requires_grad:
            np.add.at(w.grad, self.input_data.data, self.data.grad)



from ..nn.global_graph import np
import copy


class Initializer:
    def __init__(self, seed=None):
        if seed is not None:
            np.random.seed(seed)

    def decompose_size(self, size):
        size = np.array(size)
        if size.ndim == 2:
            fan_in, fan_out = size
        elif size.ndim == 4 or size.ndim == 5:
            respective_field_size = np.prod(size[2:])
            fan_in = size[1] * respective_field_size
            fan_out = size[0] * respective_field_size
        else:
            fan_in = fan_out = int(np.sqrt(np.prod(size)))
        return fan_in, fan_out


class Uniform(Initializer):
    def __init__(self, scale=0.05, seed=None):
        self.scale = scale
        super(Uniform, self).__init__(seed)

    def __call__(self, size):
        return np.random.uniform(-self.scale, self.scale, size=size)


class LecunUniform(Initializer):
    def __init__(self, seed=None):
        super(LecunUniform, self).__init__(seed)

    def __call__(self, size):
        fan_in, fan_out = self.decompose_size(size)
        return Uniform(np.sqrt(3. / fan_in))(size)


class GlorotUniform(Initializer):
    def __init__(self, seed=None):
        super(GlorotUniform, self).__init__(seed)

    def __call__(self, size):
        fan_in, fan_out = self.decompose_size(size)
        return Uniform(np.sqrt(6. / (fan_in + fan_out)))(size)


class HeUniform(Initializer):
    def __init__(self, seed=None):
        super(HeUniform, self).__init__(seed)

    def __call__(self, size):
        fan_in, fan_out = self.decompose_size(size)
        return Uniform(np.sqrt(6. / fan_in))(size)


class Normal(Initializer):
    def __init__(self, std=0.1, mean=0.0, seed=None):
        self.std = std
        self.mean = mean
        super(Normal, self).__init__(seed)

    def __call__(self, size):
        return np.random.normal(loc=self.mean, scale=self.std, size=size)


class LecunNormal(Initializer):
    def __init__(self, seed=None):
        super(LecunNormal, self).__init__(seed)

    def __call__(self, size):
        fan_in, fan_out = self.decompose_size(size)
        return Normal(np.sqrt(1. / fan_in))(size)


class GlorotNormal(Initializer):
    def __init__(self, seed=None):
        super(GlorotNormal, self).__init__(seed)

    def __call__(self, size):
        fan_in, fan_out = self.decompose_size(size)
        return Normal(np.sqrt(2. / (fan_in + fan_out)))(size)


class HeNormal(Initializer):
    def __init__(self, seed=None):
        super(HeNormal, self).__init__(seed)

    def __call__(self, size):
        fan_in, fan_out = self.decompose_size(size)
        return Normal(np.sqrt(2. / fan_in))(size)


class Orthogonal(Initializer):
    def __init__(self, gain=1.0, seed=None):
        if gain == 'relu':
            gain = np.sqrt(2)
        self.gain = gain
        super(Orthogonal, self).__init__(seed)

    def __call__(self, size):
        size = np.array(size)
        flat_shape = (size[0].tolist(), np.prod(size[1:]).tolist())
        a = Normal(1.)(flat_shape)
        u, _, v = np.linalg.svd(a, full_matrices=False)
        q = u if u.shape == flat_shape else v
        q = q.reshape(size.tolist())
        q = self.gain * q
        return q


class Zeros(Initializer):
    def __call__(self, size):
        return np.zeros(size)


class Ones(Initializer):
    def __call__(self, size):
        return np.ones(size)


def get_initializer(initializer):
    if initializer.__class__.__name__ == 'str':
        initializer = initializer.lower()
        if initializer == 'zeros':
            return Zeros()
        elif initializer == 'ones':
            return Ones()
        elif initializer in ['heuniform', 'he_uniform']:
            return HeUniform()
        elif initializer == 'uniform':
            return Uniform()
        elif initializer == 'normal':
            return Normal()
        elif initializer in ['henormal', 'he_normal']:
            return HeNormal()
        elif initializer in ['lecunnormal', 'lecun_normal']:
            return LecunNormal()
        elif initializer in ['lecununiform', 'lecun_uniform']:
            return LecunUniform()
        elif initializer =='orthogonal':
            return Orthogonal()
        elif initializer in ['glorotnoraml', 'glorot_normal']:
            return GlorotNormal()
        elif initializer in ['glorotuniform', 'glorot_uniform']:
            return GlorotUniform()

    elif isinstance(initializer, Initializer):
        return copy.deepcopy(initializer)
    else:
        raise TypeError('unknown initialization type!')



from .utils.objectives import get_objective
from .utils.optimizers import get_optimizer
from .utils.toolkit import *
from .nn.core import Variable, Layer
from .nn.functional import concatenate
from .nn.global_graph import np, topological_sort
import time
import pickle
from typing import List, Tuple, Union


class Module:
    def __init__(self, *args):
        self.trainable_variables = []  # Variable数组
        self.loss = None  # Variable
        self.optimizer = None  # Optimizer类型
        self.graph = None  # Layer数组
        self.is_training = True  # 布尔

    def __call__(self, x: Variable, *args, **kwargs):
        out = self.forward(x)
        self.__add_trainable_variables(x)
        return out

    def __add_trainable_variables(self, x: Variable):
        # 用一个列表来模拟队列，然后bfs遍历统计所有的trainable_variables
        queue = [x]
        seen = set()  # 此处为set, python里set用的是hash table, 搜索时比数组要快。
        seen.add(x)
        while queue:
            # 队列非空，队首元素出列
            vertex = queue.pop(0)
            # 将该元素的邻接元素加入队尾
            for n in vertex.out_bounds:
                if n not in seen:
                    for v in n.in_bounds:
                        if v is not None and v.requires_grad and v.name == 'variable'and v not in self.trainable_variables:
                            self.trainable_variables.append(v)

                    queue.append(n)
                    seen.add(n)

    def parameters(self):
        return self.trainable_variables

    def compile(self, optimizer, loss):
        raise NotImplemented

    def train(self):
        self.is_training = True

    def eval(self):
        self.is_training = False

    def fit(self,
            x: np.ndarray,
            y: np.ndarray,
            batch_size: int = 64,
            epochs: int = 1,
            verbose: int = 1,
            shuffle: bool = True,
            validation_data: Tuple[np.ndarray] = None,
            validation_split: float = 0.,
            initial_epoch: int = 0,
            ) -> dict:

        if validation_data is None and 0. < validation_split < 1.:
            split = int(x.shape[0] * validation_split)
            valid_x, valid_y = x[-split:], y[-split:]
            train_x, train_y = x[:-split], y[:-split]
            validation_data = (valid_x, valid_y)
        else:
            train_x, train_y = x, y

        history = dict()

        train_dataset = DataSet(train_x, train_y)
        train_dataloader = DataLoader(train_dataset, batch_size, shuffle)

        for epoch in range(initial_epoch, epochs):
            epoch_time = AverageMeter(name='epoch_time')
            train_loss = AverageMeter(name='train_loss')
            train_acc = AverageMeter(name='train_acc')
            valid_loss = AverageMeter(name='validation_loss')
            valid_acc = AverageMeter(name='validation_acc')
            history[epoch] = {
                'epoch_time': epoch_time,
                'train_loss': train_loss,
                'train_acc': train_acc,
                'validation_loss': valid_loss,
                'validation_acc': valid_acc
            }
            start_time = time.time()

            if verbose != 0:
                print('\033[1;31m Epoch[%d/%d]\033[0m' % (epoch + 1, epochs))
            progress_bar = ProgressBar(max_iter=len(train_x), verbose=verbose)

            for idx, (xs, ys) in enumerate(train_dataloader):
                xs = Variable(xs)
                ys = Variable(ys)
                self.train()
                # reset trainable_variables grad
                self.optimizer.zero_grad()
                # forward
                pred = self.forward(xs)
                loss = self.loss.forward(pred, ys)
                self.backward(loss)
                self.optimizer.step()

                epoch_time.update(time.time() - start_time)
                train_loss.update(loss.data.tolist())
                train_acc.update(self.loss.acc(pred, ys))

                if validation_data is not None:
                    valid_x, valid_y = validation_data
                    v_acc, v_loss = self.evaluate(valid_x, valid_y, batch_size=batch_size)
                    valid_loss.update(v_loss)
                    valid_acc.update(v_acc)

                progress_bar.update(xs.shape[0])
                progress_bar.console(verbose, epoch_time.val, train_loss.val, train_acc.val, valid_loss.val,
                                     valid_acc.val)

            print()
        return history

    def evaluate(self, x: np.ndarray, y: np.ndarray, batch_size: int = None):
        self.eval()
        if batch_size is not None:
            assert type(batch_size) is int
            val_dataset = DataSet(x, y)
            val_dataloader = DataLoader(val_dataset, batch_size)
            acc_list = []
            loss_list = []
            for xs, ys in val_dataloader:
                xs = Variable(xs)
                ys = Variable(ys)
                y_pred = self.forward(xs)
                metric = self.loss.metric(y_pred, ys)
                acc_list.append(metric[0])
                loss_list.append(metric[1])

            acc = np.mean(acc_list).tolist()
            loss = np.mean(loss_list).tolist()
        else:
            x = Variable(x)
            y = Variable(y)
            y_pred = self.forward(x)
            acc, loss = self.loss.metric(y_pred, y)

        return acc, loss

    def predict(self, x: np.ndarray, batch_size: int = None):
        self.eval()
        if batch_size is not None:
            assert type(batch_size) is int
            test_dataset = DataSet(x)
            test_dataloader = DataLoader(test_dataset, batch_size)
            pred_list = []
            for xs in test_dataloader:
                xs = Variable(xs)
                y_pred = self.forward(xs)
                pred_list.append(y_pred)
            pred = concatenate(*pred_list, axis=0)
        else:
            pred = self.forward(x)

        return pred

    def forward(self, x, *args):
        raise NotImplemented

    def backward(self, loss: Variable):
        loss.backward()
        for layer in reversed(self.graph):
            layer.backward()

    def save(self, save_path):
        if not save_path.endswith('.pkl'):
            save_path += '.pkl'
        with open(save_path, 'wb') as f:
            pickle.dump([self.graph, self.optimizer, self.loss], f)

    def load(self, model_path):
        if not model_path.endswith('.pkl'):
            model_path += '.pkl'
        with open(model_path, 'rb') as f:
            graph, optimizer, loss = pickle.load(f)

        self.graph = graph
        self.optimizer = optimizer
        self.loss = loss

    def __str__(self):
        bar_nums = 120
        print('*' * bar_nums)

        print('Layer(type)'.ljust(40), 'Output Shape'.ljust(40), 'Param'.ljust(12), 'Connected to'.ljust(15))
        print('#' * bar_nums)
        total_params = 0
        for layer in self.graph:
            if layer.name is not None:
                layer_name = '%s (%s)' % (layer.name, layer.__class__.__name__)
            else:
                layer_name = str(layer.__class__.__name__)

            params = layer.params_count()
            total_params += params
            first = True
            if layer.in_bounds:
                for prev_layer in layer.in_bounds:
                    if prev_layer.name is not None:
                        connected = prev_layer.name
                    else:
                        connected = prev_layer.__class__.__name__
                    if first:
                        print(layer_name.ljust(40), str((None, ) + layer.shape).ljust(40), str(params).ljust(12),
                              connected.ljust(15))
                        first = False
                    else:
                        print(''.ljust(40), ''.ljust(40), ''.ljust(12), connected.ljust(15))
            else:
                connected = '\n'
                print(layer_name.ljust(40), str((None, ) + layer.shape).ljust(40), str(params).ljust(12),
                      connected.ljust(15))
            print('-' * bar_nums)

        print('*' * bar_nums)
        trainable_params = 0
        for v in self.trainable_variables:
            trainable_params += v.data.size
        params_details = 'Total params: %d\n' % total_params
        params_details += 'Trainable params: %d\n' % trainable_params
        params_details += 'Non-trainable params: %d\n' % (total_params - trainable_params)
        return params_details


class Sequential(Module):
    def __init__(self, *layers: Layer):
        super().__init__()
        self.graph = [] if layers is None else layers

    def add(self, layer):
        self.graph.append(layer)

    def compile(self, optimizer, loss):
        assert self.graph
        next_layer = None
        for layer in self.graph:
            layer.connect(next_layer)
            layer.initial_params()
            next_layer = layer
            for v in layer.variables:
                if v is not None and v.requires_grad and v not in self.trainable_variables:
                    self.trainable_variables.append(v)
        self.loss = get_objective(loss)
        self.optimizer = get_optimizer(optimizer)
        self.optimizer.trainable_variables = self.trainable_variables

    def forward(self, x, *args):
        for layer in self.graph:
            if hasattr(layer, 'variables') and len(layer.variables) == 0:
                layer.initial_params(x.shape[1:])
            x = layer.forward(x, self.is_training)
        return x

    def pop(self, index=-1):
        layer = self.graph.pop(index)
        del layer
        print('success delete %s layer' % layer.__class__.__name__)

    def save(self, save_path):
        with open(save_path + '.pkl', 'wb') as f:
            pickle.dump([self.graph, self.optimizer, self.loss], f)

    def load(self, model_path):
        with open(model_path + '.pkl', 'rb') as f:
            layers, optimizer, loss = pickle.load(f)

        self.graph = layers
        self.optimizer = optimizer
        self.loss = loss


class Model(Module):
    def __init__(self, inputs=None, outputs=None):
        super().__init__()
        self.inputs = inputs
        self.outputs = outputs

    def compile(self, optimizer, loss, **kwargs):
        assert self.inputs is not None and self.outputs is not None
        self.graph = topological_sort(self.inputs, self.outputs)
        for g in self.graph:
            g.initial_params()
            for v in g.variables:
                if v is not None and v.requires_grad and v not in self.trainable_variables:
                    self.trainable_variables.append(v)
        self.loss = get_objective(loss)
        self.optimizer = get_optimizer(optimizer, **kwargs)
        self.optimizer.trainable_variables = self.trainable_variables

    def forward(self, x: Variable, *args):
        self.inputs.input_data = x
        outputs = None
        for layer in self.graph:
            outputs = layer.forward(is_training=self.is_training)
        return outputs

